# Copyright (C) 2021-2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import math
from enum import IntEnum
from functools import partial
from itertools import zip_longest
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import attr
import numpy as np
import shapely.geometry as sg
from attr import asdict, attrs, field
from typing_extensions import Literal

from datumaro.components.media import Image
from datumaro.util.attrs_util import default_if_none, not_empty


class AnnotationType(IntEnum):
    unknown = 0
    label = 1
    mask = 2
    points = 3
    polygon = 4
    polyline = 5
    bbox = 6
    caption = 7
    cuboid_3d = 8
    super_resolution_annotation = 9
    depth_annotation = 10
    ellipse = 11
    hash_key = 12
    feature_vector = 13
    tabular = 14
    rotated_bbox = 15


COORDINATE_ROUNDING_DIGITS = 2
CHECK_POLYGON_EQ_EPSILONE = 1e-7
NO_GROUP = 0
NO_OBJECT_ID = -1


@attrs(slots=True, kw_only=True, order=False)
class Annotation:
    """
    A base annotation class.

    Derived classes must define the '_type' class variable with a value
    from the AnnotationType enum.
    """

    # Describes an identifier of the annotation
    # Is not required to be unique within DatasetItem annotations or dataset
    id: int = field(default=0, validator=default_if_none(int))

    # Arbitrary annotation-specific attributes. Typically, includes
    # metainfo and properties that are not covered by other fields.
    # If possible, try to limit value types of values by the simple
    # builtin types (int, float, bool, str) to increase compatibility with
    # different formats.
    # There are some established names for common attributes like:
    # - "occluded" (bool)
    # - "visible" (bool)
    # Possible dataset attributes can be described in Categories.attributes.
    attributes: Dict[str, Any] = field(factory=dict, validator=default_if_none(dict))

    # Annotations can be grouped, which means they describe parts of a
    # single object. The value of 0 means there is no group.
    group: int = field(default=NO_GROUP, validator=default_if_none(int))

    # obeject identifier over the multiple items
    # e.g.) in a video, person 'A' could be annotated on the multiple frame images
    #   the user could assign >=0 value as id of person 'A'.
    object_id: int = field(default=NO_OBJECT_ID, validator=default_if_none(int))

    _type = AnnotationType.unknown

    @property
    def type(self) -> AnnotationType:
        return self._type  # must be set in subclasses

    def as_dict(self) -> Dict[str, Any]:
        "Returns a dictionary { field_name: value }"
        return asdict(self)

    def wrap(self, **kwargs):
        "Returns a modified copy of the object"
        return attr.evolve(self, **kwargs)


@attrs(slots=True, kw_only=True, order=False)
class Categories:
    """
    A base class for annotation metainfo. It is supposed to include
    dataset-wide metainfo like available labels, label colors,
    label attributes etc.
    """

    # Describes the list of possible annotation-type specific attributes
    # in a dataset.
    attributes: Set[str] = field(factory=set, validator=default_if_none(set), eq=False)


class GroupType(IntEnum):
    EXCLUSIVE = 0
    INCLUSIVE = 1
    RESTRICTED = 2

    def to_str(self) -> str:
        return self.name.lower()

    @classmethod
    def from_str(cls, text: str) -> GroupType:
        try:
            return cls[text.upper()]
        except KeyError:
            raise ValueError(f"Invalid GroupType: {text}")


@attrs(slots=True, order=False)
class LabelCategories(Categories):
    @attrs(slots=True, order=False)
    class Category:
        name: str = field(converter=str, validator=not_empty)
        parent: str = field(default="", validator=default_if_none(str))
        attributes: Set[str] = field(factory=set, validator=default_if_none(set))

    @attrs(slots=True, order=False)
    class LabelGroup:
        name: str = field(converter=str, validator=not_empty)
        labels: List[str] = field(default=[], validator=default_if_none(list))
        group_type: GroupType = field(
            default=GroupType.EXCLUSIVE, validator=default_if_none(GroupType)
        )

    items: List[str] = field(factory=list, validator=default_if_none(list))
    label_groups: List[LabelGroup] = field(factory=list, validator=default_if_none(list))
    _indices: Dict[str, int] = field(factory=dict, init=False, eq=False)

    @classmethod
    def from_iterable(
        cls,
        iterable: Iterable[
            Union[
                str,
                Tuple[str],
                Tuple[str, str],
                Tuple[str, str, List[str]],
            ]
        ],
    ) -> LabelCategories:
        """
        Creates a LabelCategories from iterable.

        Args:
            iterable: This iterable object can be:

                - a list of str - will be interpreted as list of Category names
                - a list of positional arguments - will generate Categories
                  with these arguments

        Returns: a LabelCategories object
        """

        temp_categories = cls()

        for category in iterable:
            if isinstance(category, str):
                category = [category]
            temp_categories.add(*category)

        return temp_categories

    def __attrs_post_init__(self):
        self._reindex()

    def _reindex(self):
        indices = {}
        for index, item in enumerate(self.items):
            assert item.name not in self._indices
            indices[item.name] = index
        self._indices = indices

    def add(
        self,
        name: str,
        parent: Optional[str] = None,
        attributes: Optional[Set[str]] = None,
    ) -> int:
        assert name
        assert name not in self._indices, name

        index = len(self.items)
        self.items.append(self.Category(name, parent, attributes))
        self._indices[name] = index
        return index

    def add_label_group(
        self,
        name: str,
        labels: List[str],
        group_type: GroupType,
    ) -> int:
        assert name

        index = len(self.label_groups)
        self.label_groups.append(self.LabelGroup(name, labels, group_type))
        return index

    def find(self, name: str) -> Tuple[Optional[int], Optional[Category]]:
        index = self._indices.get(name)
        if index is not None:
            return index, self.items[index]
        return index, None

    def __getitem__(self, idx: int) -> Category:
        return self.items[idx]

    def __contains__(self, value: Union[int, str]) -> bool:
        if isinstance(value, str):
            return self.find(value)[1] is not None
        else:
            return 0 <= value and value < len(self.items)

    def __len__(self) -> int:
        return len(self.items)

    def __iter__(self) -> Iterator[Category]:
        return iter(self.items)


@attrs(slots=True, order=False)
class Label(Annotation):
    _type = AnnotationType.label
    label: int = field(converter=int)


@attrs(slots=True, eq=False, order=False)
class HashKey(Annotation):
    _type = AnnotationType.hash_key
    hash_key: np.ndarray = field(validator=attr.validators.instance_of(np.ndarray))

    @hash_key.validator
    def _validate(self, attribute, value: np.ndarray):
        """Check whether value is a 1D Numpy array having 64 np.uint8 values"""
        if value.ndim != 1 or value.shape[0] != 64 or value.dtype != np.uint8:
            raise ValueError(value)

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        if not isinstance(other, __class__):
            return False
        return np.array_equal(self.hash_key, other.hash_key)


@attrs(eq=False, order=False)
class FeatureVector(Annotation):
    _type = AnnotationType.feature_vector
    vector: np.ndarray = field(validator=attr.validators.instance_of(np.ndarray))

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        if not isinstance(other, __class__):
            return False
        return np.array_equal(self.hash_key, other.hash_key)


RgbColor = Tuple[int, int, int]

Colormap = Dict[int, RgbColor]
"""Represents { index -> color } mapping for segmentation masks"""


@attrs(slots=True, eq=False, order=False)
class MaskCategories(Categories):
    """
    Describes a color map for segmentation masks.
    """

    @classmethod
    def generate(cls, size: int = 255, include_background: bool = True) -> MaskCategories:
        """
        Generates MaskCategories with the specified size.

        If include_background is True, the result will include the item
            "0: (0, 0, 0)", which is typically used as a background color.
        """
        from datumaro.util.mask_tools import generate_colormap

        return cls(generate_colormap(size, include_background=include_background))

    colormap: Colormap = field(factory=dict, validator=default_if_none(dict))
    _inverse_colormap: Optional[Dict[RgbColor, int]] = field(
        default=None, validator=attr.validators.optional(dict)
    )

    @property
    def inverse_colormap(self) -> Dict[RgbColor, int]:
        from datumaro.util.mask_tools import invert_colormap

        if self._inverse_colormap is None:
            if self.colormap is not None:
                self._inverse_colormap = invert_colormap(self.colormap)
        return self._inverse_colormap

    def __contains__(self, idx: int) -> bool:
        return idx in self.colormap

    def __getitem__(self, idx: int) -> RgbColor:
        return self.colormap[idx]

    def __len__(self) -> int:
        return len(self.colormap)

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        if not isinstance(other, __class__):
            return False
        for label_id, my_color in self.colormap.items():
            other_color = other.colormap.get(label_id)
            if not np.array_equal(my_color, other_color):
                return False
        return True


BinaryMaskImage = np.ndarray  # 2d array of type bool
BinaryMaskImageCallable = Callable[[], BinaryMaskImage]
IndexMaskImage = np.ndarray  # 2d array of type int
IndexMaskImageCallable = Callable[[], IndexMaskImage]


@attrs(slots=True, eq=False, order=False)
class Mask(Annotation):
    """
    Represents a 2d single-instance binary segmentation mask.
    """

    _type = AnnotationType.mask
    _image: Union[BinaryMaskImage, BinaryMaskImageCallable] = field()
    label: Optional[int] = field(
        converter=attr.converters.optional(int), default=None, kw_only=True
    )
    z_order: int = field(default=0, validator=default_if_none(int), kw_only=True)

    def __attrs_post_init__(self):
        if isinstance(self._image, np.ndarray):
            self._image = self._image.astype(bool)

    @property
    def image(self) -> BinaryMaskImage:
        image = self._image
        if callable(image):
            image = image()
        return image

    def as_class_mask(
        self,
        label_id: Optional[int] = None,
        ignore_index: int = 0,
        dtype: Optional[np.dtype] = None,
    ) -> IndexMaskImage:
        """Produces a class index mask based on the binary mask.

        Args:
            label_id: Scalar value to represent the class index of the mask.
                If not specified, `self.label` will be used. Defaults to None.
            ignore_index: Scalar value to fill in the zeros in the binary mask.
                Defaults to 0.
            dtype: Data type for the resulting mask. If not specified,
                it will be inferred from the provided `label_id` to hold its value.
                For example, if `label_id=255`, the inferred dtype will be `np.uint8`.
                Defaults to None.

        Returns:
            IndexMaskImage: Class index mask generated from the binary mask.
        """
        if label_id is None:
            label_id = self.label
        from datumaro.util.mask_tools import make_index_mask

        return make_index_mask(self.image, index=label_id, ignore_index=ignore_index, dtype=dtype)

    def as_instance_mask(
        self,
        instance_id: int,
        ignore_index: int = 0,
        dtype: Optional[np.dtype] = None,
    ) -> IndexMaskImage:
        """Produces an instance index mask based on the binary mask.

        Args:
            instance_id: Scalar value to represent the instance id.
            ignore_index: Scalar value to fill in the zeros in the binary mask.
                Defaults to 0.
            dtype: Data type for the resulting mask. If not specified,
                it will be inferred from the provided `label_id` to hold its value.
                For example, if `label_id=255`, the inferred dtype will be `np.uint8`.
                Defaults to None.

        Returns:
            IndexMaskImage: Instance index mask generated from the binary mask.
        """
        from datumaro.util.mask_tools import make_index_mask

        return make_index_mask(
            self.image, index=instance_id, ignore_index=ignore_index, dtype=dtype
        )

    def get_area(self) -> int:
        return np.count_nonzero(self.image)

    def get_bbox(self) -> Tuple[int, int, int, int]:
        """
        Computes the bounding box of the mask.

        Returns: [x, y, w, h]
        """
        from datumaro.util.mask_tools import find_mask_bbox

        return find_mask_bbox(self.image)

    def paint(self, colormap: Colormap) -> np.ndarray:
        """
        Applies a colormap to the mask and produces the resulting image.
        """
        from datumaro.util.mask_tools import paint_mask

        return paint_mask(self.as_class_mask(), colormap)

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        if not isinstance(other, __class__):
            return False
        return (
            (self.label == other.label)
            and (self.z_order == other.z_order)
            and (np.array_equal(self.image, other.image))
        )


@attrs(slots=True, eq=False, order=False)
class RleMask(Mask):
    """
    An RLE-encoded instance segmentation mask.
    """

    _rle = field()  # uses pycocotools RLE representation

    _image = field(init=False, default=None)

    @property
    def image(self) -> BinaryMaskImage:
        return self._decode(self.rle)

    @property
    def rle(self):
        rle = self._rle
        if callable(rle):
            rle = rle()
        return rle

    @staticmethod
    def _decode(rle):
        from pycocotools import mask as mask_utils

        return mask_utils.decode(rle)

    def get_area(self) -> int:
        from pycocotools import mask as mask_utils

        return mask_utils.area(self.rle)

    def get_bbox(self) -> Tuple[int, int, int, int]:
        from pycocotools import mask as mask_utils

        return mask_utils.toBbox(self.rle)

    def __eq__(self, other):
        if not isinstance(other, __class__):
            return super().__eq__(other)
        return self.rle == other.rle


@attrs(slots=True, eq=False, order=False)
class ExtractedMask(Mask):
    """Mask annotation (binary mask) extracted from an index mask (integer 2D Numpy array).

    This class can extract a binary mask with given index mask and index value.
    The advantage of this class is that we can create multiple binary mask but they share a single index mask source.

    Attributes:
        index_mask: Integer 2D Numpy array. Its pixel can indicate a label id (class) or an instance id.
        index: Integer value to extract a binary mask from the given index mask.

    Examples:
        This example demonstrates how to create an `ExtractedMask` from a synthetic index mask,
        which denotes a semantic segmentation mask with binary values such as 0 for background
        and 1 for foreground.

        >>> import numpy as np
        >>> from datumaro.components.annotation import ExtractedMask
        >>>
        >>> index_mask = np.random.randint(low=0, high=2, size=(10, 10), dtype=np.uint8)
        >>> mask1 = ExtractedMask(index_mask=index_mask, index=0, label=0)  # 0 for background
        >>> mask2 = ExtractedMask(index_mask=index_mask, index=1, label=1)  # 1 for foreground
        >>> np.unique(mask1.image).tolist()  # `image` property create a binary mask
        np.array([0, 1])
        >>> mask1.index_mask == mask2.index_mask  # They share the same source
        True
    """

    index_mask: Union[IndexMaskImage, IndexMaskImageCallable] = field()
    index: int = field()

    _image: None = field(init=False, default=None)

    @property
    def image(self) -> BinaryMaskImage:
        index_mask = self.index_mask() if callable(self.index_mask) else self.index_mask
        return index_mask == self.index


CompiledMaskImage = np.ndarray  # 2d of integers (of different precision)


class CompiledMask:
    """
    Represents class- and instance- segmentation masks with
    all the instances (opposed to single-instance masks).
    """

    @staticmethod
    def from_instance_masks(
        instance_masks: Iterable[Mask],
        instance_ids: Optional[Iterable[int]] = None,
        instance_labels: Optional[Iterable[int]] = None,
    ) -> CompiledMask:
        """
        Joins instance masks into a single mask. Masks are sorted by
        z_order (ascending) prior to merging.

        Parameters:
            instance_ids: Instance id values for the produced instance mask.
                By default, mask positions are used.
            instance_labels: Instance label id values for the produced class
                mask. By default, mask labels are used.
        """

        from datumaro.util.mask_tools import make_index_mask

        instance_ids = instance_ids or []
        instance_labels = instance_labels or []
        masks = sorted(
            zip_longest(instance_masks, instance_ids, instance_labels), key=lambda m: m[0].z_order
        )

        max_index = len(masks) + 1
        index_dtype = np.min_scalar_type(max_index)

        masks = (
            (m, 1 + i, id if id is not None else 1 + i, label if label is not None else m.label)
            for i, (m, id, label) in enumerate(masks)
        )

        # This optimized version is supposed for:
        # 1. Avoiding memory explosion on materialization of all masks
        # 2. Optimizing mask materialization calls (RLE decoding)
        # 3. Optimizing intermediate mask memory use
        #
        # Basically, a mask can be quite large (e.g. 10k x 10k @ int32 etc.),
        # so we can only afford having just few copies in
        # memory simultaneously.

        it = iter(masks)

        instance_map = [0]
        class_map = [0]

        m, idx, instance_id, class_id = next(it)
        if not class_id:
            idx = 0
        index_mask = make_index_mask(m.image, idx, dtype=index_dtype)
        instance_map.append(instance_id)
        class_map.append(class_id)

        for m, idx, instance_id, class_id in it:
            if not class_id:
                idx = 0
            index_mask = np.where(m.image, idx, index_mask)
            instance_map.append(instance_id)
            class_map.append(class_id)

        # Generate compiled masks

        if np.array_equal(instance_map, range(max_index)):
            merged_instance_mask = index_mask
        else:
            merged_instance_mask = np.array(instance_map, dtype=np.min_scalar_type(instance_map))[
                index_mask
            ]

        merged_class_mask = np.array(class_map, dtype=np.min_scalar_type(class_map))[index_mask]

        return __class__(class_mask=merged_class_mask, instance_mask=merged_instance_mask)

    def __init__(
        self,
        class_mask: Union[None, CompiledMaskImage, Callable[[], CompiledMaskImage]] = None,
        instance_mask: Union[None, CompiledMaskImage, Callable[[], CompiledMaskImage]] = None,
    ):
        self._class_mask = class_mask
        self._instance_mask = instance_mask

    @staticmethod
    def _get_image(image):
        if callable(image):
            return image()
        return image

    @property
    def class_mask(self) -> Optional[CompiledMaskImage]:
        return self._get_image(self._class_mask)

    @property
    def instance_mask(self) -> Optional[CompiledMaskImage]:
        return self._get_image(self._instance_mask)

    @property
    def instance_count(self) -> int:
        return int(self.instance_mask.max())

    def get_instance_labels(self) -> Dict[int, int]:
        """
        Matches the class and instance masks.

        Returns: { instance id: class id }
        """

        class_shift = 16
        m = (self.class_mask.astype(np.uint32) << class_shift) + self.instance_mask.astype(
            np.uint32
        )
        keys = np.unique(m)
        instance_labels = {
            int(k & ((1 << class_shift) - 1)): int(k >> class_shift)
            for k in keys
            if k & ((1 << class_shift) - 1) != 0
        }
        return instance_labels

    def extract(self, instance_id: int) -> IndexMaskImage:
        """
        Extracts a single-instance mask from the compiled mask.
        """

        return self.instance_mask == instance_id

    def lazy_extract(self, instance_id: int) -> Callable[[], IndexMaskImage]:
        return partial(self.extract, instance_id)


@attrs(slots=True, order=False)
class Shape(Annotation):
    """
    Base class for shape annotations. This class defines the common attributes and methods
    for different types of shape annotations.

    Attributes:
        points (List[float]): List of float values representing the coordinates of the shape.
        label (Optional[int]): Optional label ID for the shape. Default is None.
        z_order (int): Z-order of the shape, used to determine the rendering order. Default is 0.

    Methods:
        get_area: Abstract method to calculate the area of the shape.
        as_polygon: Abstract method to convert the shape into a polygon representation.
        get_bbox: Returns the bounding box of the shape as [x, y, w, h].
        get_points: Returns the points of the shape as a list of (x, y) tuples.
    """

    points: List[float] = field(
        converter=lambda x: np.array(x, dtype=np.float32).round(COORDINATE_ROUNDING_DIGITS).tolist()
    )

    label: Optional[int] = field(
        converter=attr.converters.optional(int), default=None, kw_only=True
    )

    z_order: int = field(default=0, validator=default_if_none(int), kw_only=True)

    def get_area(self):
        """
        Calculate the area of the shape.
        """
        raise NotImplementedError()

    def as_polygon(self) -> List[float]:
        """
        Convert the shape into a polygon representation.
        """
        raise NotImplementedError()

    def get_bbox(self) -> Tuple[float, float, float, float]:
        """
        Calculate and return the bounding box of the shape.

        Returns:
            Tuple[float, float, float, float]: The bounding box as [x, y, w, h].
        """

        points = self.points
        if not points:
            return None

        xs = [p for p in points[0::2]]
        ys = [p for p in points[1::2]]
        x0 = min(xs)
        x1 = max(xs)
        y0 = min(ys)
        y1 = max(ys)
        return [x0, y0, x1 - x0, y1 - y0]

    def get_points(self) -> Optional[List[Tuple[float, float]]]:
        """
        Convert and return the points of the shape as a list of (x, y) tuples.

        Returns:
            Optional[List[Tuple[float, float]]]: List of points as (x, y) tuples, or None if no points.
        """
        points = self.points
        if not points:
            return None

        assert len(points) % 2 == 0, "points should have (2 x points) number of float values."

        xs = [p for p in points[0::2]]
        ys = [p for p in points[1::2]]

        return [(x, y) for x, y in zip(xs, ys)]


@attrs(slots=True, order=False)
class PolyLine(Shape):
    """
    PolyLine annotation class. This class represents a polyline shape, which is a series of connected line segments.

    Attributes:
        _type (AnnotationType): The type of annotation, set to `AnnotationType.polyline`.

    Methods:
        as_polygon: Returns the points of the polyline as a polygon.
        get_area: Returns the area of the polyline, which is always 0.
    """

    _type = AnnotationType.polyline

    def as_polygon(self):
        return self.points[:]

    def get_area(self):
        return 0


@attrs(slots=True, init=False, order=False)
class Cuboid3d(Annotation):
    """
    Cuboid3d annotation class. This class represents a 3D cuboid annotation with position, rotation, and scale.

    Attributes:
        _type (AnnotationType): The type of annotation, set to `AnnotationType.cuboid_3d`.
        _points (List[float]): List of float values representing the position, rotation, and scale of the cuboid.
        label (Optional[int]): Optional label ID for the cuboid. Default is None.

    Methods:
        __init__: Initializes the Cuboid3d with position, rotation, and scale.
        position: Property to get and set the position of the cuboid.
        rotation: Property to get and set the rotation of the cuboid.
        scale: Property to get and set the scale of the cuboid.
    """

    _type = AnnotationType.cuboid_3d
    _points: List[float] = field(default=None)
    label: Optional[int] = field(
        converter=attr.converters.optional(int), default=None, kw_only=True
    )

    @_points.validator
    def _points_validator(self, attribute, points):
        """
        Validate and round the points representing the cuboid's position, rotation, and scale.

        Args:
            attribute: The attribute being validated.
            points: The list of float values to validate.
        """
        if points is None:
            points = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0]
        else:
            assert len(points) == 3 + 3 + 3, points
            points = np.around(points, COORDINATE_ROUNDING_DIGITS).tolist()
        self._points = points

    def __init__(self, position, rotation=None, scale=None, **kwargs):
        """
        Initialize the Cuboid3d with position, rotation, and scale.

        Args:
            position (List[float]): List of 3 float values representing the position [x, y, z].
            rotation (List[float], optional): List of 3 float values representing the rotation [rx, ry, rz].
            scale (List[float], optional): List of 3 float values representing the scale [sx, sy, sz].
        """
        assert len(position) == 3, position
        if not rotation:
            rotation = [0] * 3
        if not scale:
            scale = [1] * 3
        kwargs.pop("points", None)
        self.__attrs_init__(points=[*position, *rotation, *scale], **kwargs)

    @property
    def position(self):
        """
        Get the position of the cuboid.

        Returns:
            List[float]: The position [x, y, z] of the cuboid.
        """
        return self._points[0:3]

    @position.setter
    def _set_poistion(self, value):
        """
        Set the position of the cuboid.

        Args:
            value (List[float]): The new position [x, y, z] of the cuboid.
        """
        # TODO: fix the issue with separate coordinate rounding:
        # self.position[0] = 12.345676
        # - the number assigned won't be rounded.
        self.position[:] = np.around(value, COORDINATE_ROUNDING_DIGITS).tolist()

    @property
    def rotation(self):
        """
        Get the rotation of the cuboid.

        Returns:
            List[float]: The rotation [rx, ry, rz] of the cuboid.
        """
        return self._points[3:6]

    @rotation.setter
    def _set_rotation(self, value):
        """
        Set the rotation of the cuboid.

        Args:
            value (List[float]): The new rotation [rx, ry, rz] of the cuboid.
        """
        self.rotation[:] = np.around(value, COORDINATE_ROUNDING_DIGITS).tolist()

    @property
    def scale(self):
        """
        Get the scale of the cuboid.

        Returns:
            List[float]: The scale [sx, sy, sz] of the cuboid.
        """
        return self._points[6:9]

    @scale.setter
    def _set_scale(self, value):
        """
        Set the scale of the cuboid.

        Args:
            value (List[float]): The new scale [sx, sy, sz] of the cuboid.
        """
        self.scale[:] = np.around(value, COORDINATE_ROUNDING_DIGITS).tolist()


@attrs(slots=True, order=False, eq=False)
class Polygon(Shape):
    """
    Polygon annotation class. This class represents a polygon shape defined by a series of points.

    Attributes:
        _type (AnnotationType): The type of annotation, set to `AnnotationType.polygon`.

    Methods:
        __attrs_post_init__: Validates the points to ensure they form a valid polygon.
        get_area: Calculates the area of the polygon using the shoelace formula.
        as_polygon: Returns the points of the polygon.
        __eq__: Compares this polygon with another for equality.
        _get_shoelace_area: Helper method to calculate the area of the polygon using the shoelace formula.
    """

    _type = AnnotationType.polygon

    def __attrs_post_init__(self):
        """
        Validate the points to ensure they form a valid polygon.

        Raises:
            AssertionError: If the number of points is not even or less than 3 pairs of coordinates.
        """
        # keep the message on a single line to produce informative output
        assert len(self.points) % 2 == 0 and 3 <= len(self.points) // 2, (
            "Wrong polygon points: %s" % self.points
        )

    def get_area(self):
        """
        Calculate the area of the polygon using the shoelace formula.

        Returns:
            float: The area of the polygon.
        """
        # import pycocotools.mask as mask_utils

        # x, y, w, h = self.get_bbox()
        # rle = mask_utils.frPyObjects([self.points], y + h, x + w)
        # area = mask_utils.area(rle)[0]
        area = self._get_shoelace_area()
        return area

    def as_polygon(self) -> List[float]:
        """
        Return the points of the polygon.

        Returns:
            List[float]: The points of the polygon.
        """
        return self.points

    def __eq__(self, other):
        """
        Compare this polygon with another for equality.

        Args:
            other: The other polygon to compare with.

        Returns:
            bool: True if the polygons are equal, False otherwise.
        """
        if not isinstance(other, __class__):
            return False
        if (
            not Annotation.__eq__(self, other)
            or self.label != other.label
            or self.z_order != other.z_order
        ):
            return False

        self_points = self.get_points()
        other_points = other.get_points()
        self_polygon = sg.Polygon(self_points)
        other_polygon = sg.Polygon(other_points)
        # if polygon is not valid, compare points
        if not (self_polygon.is_valid and other_polygon.is_valid):
            return self_points == other_points
        inter_area = self_polygon.intersection(other_polygon).area
        return abs(self_polygon.area - inter_area) < CHECK_POLYGON_EQ_EPSILONE

    def _get_shoelace_area(self):
        """
        Calculate the area of the polygon using the shoelace formula.

        Returns:
            float: The area of the polygon.
        """
        points = self.get_points()
        n = len(points)
        # Not a polygon
        if n < 3:
            return 0

        area = 0.0
        for i in range(n):
            x1, y1 = points[i]
            x2, y2 = points[(i + 1) % n]  # Next vertex, wrapping around using modulo
            area += x1 * y2 - y1 * x2

        return abs(area) / 2.0


@attrs(slots=True, init=False, order=False)
class Bbox(Shape):
    """
    Bbox annotation class. This class represents a bounding box defined by its top-left corner (x, y)
    and its width and height (w, h).

    Attributes:
        _type (AnnotationType): The type of annotation, set to `AnnotationType.bbox`.

    Methods:
        __init__: Initializes the Bbox with its coordinates and dimensions.
        x: Property to get the x-coordinate of the bounding box.
        y: Property to get the y-coordinate of the bounding box.
        w: Property to get the width of the bounding box.
        h: Property to get the height of the bounding box.
        get_area: Calculates the area of the bounding box.
        get_bbox: Returns the bounding box coordinates and dimensions.
        as_polygon: Returns the bounding box as a list of points forming a polygon.
        iou: Calculates the Intersection over Union (IoU) with another shape.
        wrap: Creates a new Bbox instance with updated attributes.
    """

    _type = AnnotationType.bbox

    def __init__(self, x, y, w, h, *args, **kwargs):
        """
        Initialize the Bbox with its top-left corner (x, y) and its width and height (w, h).

        Args:
            x (float): The x-coordinate of the top-left corner.
            y (float): The y-coordinate of the top-left corner.
            w (float): The width of the bounding box.
            h (float): The height of the bounding box.
        """
        kwargs.pop("points", None)  # comes from wrap()
        self.__attrs_init__([x, y, x + w, y + h], *args, **kwargs)

    @property
    def x(self):
        """
        Get the x-coordinate of the top-left corner of the bounding box.

        Returns:
            float: The x-coordinate of the bounding box.
        """
        return self.points[0]

    @property
    def y(self):
        """
        Get the y-coordinate of the top-left corner of the bounding box.

        Returns:
            float: The y-coordinate of the bounding box.
        """
        return self.points[1]

    @property
    def w(self):
        """
        Get the width of the bounding box.

        Returns:
            float: The width of the bounding box.
        """
        return self.points[2] - self.points[0]

    @property
    def h(self):
        """
        Get the height of the bounding box.

        Returns:
            float: The height of the bounding box.
        """
        return self.points[3] - self.points[1]

    def get_area(self):
        """
        Calculate the area of the bounding box.

        Returns:
            float: The area of the bounding box.
        """
        return self.w * self.h

    def get_bbox(self):
        """
        Get the bounding box coordinates and dimensions.

        Returns:
            List[float]: The bounding box as [x, y, w, h].
        """
        return [self.x, self.y, self.w, self.h]

    def as_polygon(self) -> List[float]:
        """
        Convert the bounding box into a polygon representation.

        Returns:
            List[float]: The bounding box as a polygon.
        """
        x, y, w, h = self.get_bbox()
        return [x, y, x + w, y, x + w, y + h, x, y + h]

    def iou(self, other: Shape) -> Union[float, Literal[-1]]:
        """
        Calculate the Intersection over Union (IoU) with another shape.

        Args:
            other (Shape): The other shape to compare with.

        Returns:
            Union[float, Literal[-1]]: The IoU value or -1 if not applicable.
        """
        from datumaro.util.annotation_util import bbox_iou

        return bbox_iou(self.get_bbox(), other.get_bbox())

    def wrap(item, **kwargs):
        """
        Create a new Bbox instance with updated attributes.

        Args:
            item (Bbox): The original Bbox instance.
            kwargs: Additional attributes to update.

        Returns:
            Bbox: A new Bbox instance with updated attributes.
        """
        d = {"x": item.x, "y": item.y, "w": item.w, "h": item.h}
        d.update(kwargs)
        return attr.evolve(item, **d)


@attrs(slots=True, init=False, order=False)
class RotatedBbox(Shape):
    """
    RotatedBbox annotation class. This class represents a rotated bounding box defined
    by its center (cx, cy), width (w), height (h), and rotation angle (r).

    Attributes:
        _type (AnnotationType): The type of annotation, set to `AnnotationType.rotated_bbox`.

    Methods:
        __init__: Initializes the RotatedBbox with its center, dimensions, and rotation angle.
        from_rectangle: Creates a RotatedBbox from a list of four corner points.
        cx: Property to get the x-coordinate of the center of the bounding box.
        cy: Property to get the y-coordinate of the center of the bounding box.
        w: Property to get the width of the bounding box.
        h: Property to get the height of the bounding box.
        r: Property to get the rotation angle of the bounding box.
        get_area: Calculates the area of the bounding box.
        get_bbox: Returns the bounding box coordinates and dimensions.
        get_rotated_bbox: Returns the rotated bounding box parameters.
        as_polygon: Converts the rotated bounding box into a list of corner points.
        iou: Calculates the Intersection over Union (IoU) with another shape.
        wrap: Creates a new RotatedBbox instance with updated attributes.
    """

    _type = AnnotationType.rotated_bbox

    def __init__(self, cx, cy, w, h, r, *args, **kwargs):
        """
        Initialize the RotatedBbox with its center (cx, cy), width (w), height (h), and rotation angle (r).

        Args:
            cx (float): The x-coordinate of the center.
            cy (float): The y-coordinate of the center.
            w (float): The width of the bounding box.
            h (float): The height of the bounding box.
            r (float): The rotation angle of the bounding box in degrees.
        """
        kwargs.pop("points", None)  # comes from wrap()
        self.__attrs_init__([cx, cy, w, h, r], *args, **kwargs)

    @classmethod
    def from_rectangle(cls, points: List[Tuple[float, float]], *args, **kwargs):
        """
        Create a RotatedBbox from a list of four corner points.

        Args:
            points (List[Tuple[float, float]]): A list of four points defining the rectangle.
            args: Additional arguments.
            kwargs: Additional keyword arguments.

        Returns:
            RotatedBbox: A new RotatedBbox instance.
        """
        assert len(points) == 4, "polygon for a rotated bbox should have only 4 coordinates."

        # Calculate rotation angle
        rot = math.atan2(points[1][1] - points[0][1], points[1][0] - points[0][0])

        # Calculate the center of the bounding box
        cx = (points[0][0] + points[2][0]) / 2
        cy = (points[0][1] + points[2][1]) / 2

        # Calculate the width and height
        width = math.sqrt((points[1][0] - points[0][0]) ** 2 + (points[1][1] - points[0][1]) ** 2)
        height = math.sqrt((points[2][0] - points[1][0]) ** 2 + (points[2][1] - points[1][1]) ** 2)

        return cls(cx=cx, cy=cy, w=width, h=height, r=math.degrees(rot), *args, **kwargs)

    @property
    def cx(self):
        """
        Get the x-coordinate of the center of the bounding box.

        Returns:
            float: The x-coordinate of the center.
        """
        return self.points[0]

    @property
    def cy(self):
        """
        Get the y-coordinate of the center of the bounding box.

        Returns:
            float: The y-coordinate of the center.
        """
        return self.points[1]

    @property
    def w(self):
        """
        Get the width of the bounding box.

        Returns:
            float: The width of the bounding box.
        """
        return self.points[2]

    @property
    def h(self):
        """
        Get the height of the bounding box.

        Returns:
            float: The height of the bounding box.
        """
        return self.points[3]

    @property
    def r(self):
        """
        Get the rotation angle of the bounding box in degrees.

        Returns:
            float: The rotation angle of the bounding box.
        """
        return self.points[4]

    def get_area(self):
        """
        Calculate the area of the bounding box.

        Returns:
            float: The area of the bounding box.
        """
        return self.w * self.h

    def get_bbox(self):
        """
        Get the bounding box coordinates and dimensions.

        Returns:
            List[float]: The bounding box as [x, y, w, h].
        """
        polygon = self.as_polygon()
        xs = [pt[0] for pt in polygon]
        ys = [pt[1] for pt in polygon]

        return [min(xs), min(ys), max(xs) - min(xs), max(ys) - min(ys)]

    def get_rotated_bbox(self):
        """
        Get the rotated bounding box parameters.

        Returns:
            List[float]: The rotated bounding box as [cx, cy, w, h, r].
        """
        return [self.cx, self.cy, self.w, self.h, self.r]

    def as_polygon(self) -> List[Tuple[float, float]]:
        """
        Convert the rotated bounding box into a list of corner points.

        Returns:
            List[Tuple[float, float]]: The bounding box as a list of four corner points.
        """

        def _rotate_point(x, y, angle):
            """
            Rotate a point around another point.

            Args:
                x (float): The x-coordinate of the point.
                y (float): The y-coordinate of the point.
                angle (float): The rotation angle in degrees.

            Returns:
                Tuple[float, float]: The rotated point coordinates.
            """
            angle_rad = math.radians(angle)
            cos_theta = math.cos(angle_rad)
            sin_theta = math.sin(angle_rad)
            nx = cos_theta * x - sin_theta * y
            ny = sin_theta * x + cos_theta * y
            return nx, ny

        # Calculate corner points of the rectangle
        corners = [
            (-self.w / 2, -self.h / 2),
            (self.w / 2, -self.h / 2),
            (self.w / 2, self.h / 2),
            (-self.w / 2, self.h / 2),
        ]

        # Rotate each corner point
        rotated_corners = [_rotate_point(p[0], p[1], self.r) for p in corners]

        # Translate the rotated points to the original position
        return [(p[0] + self.cx, p[1] + self.cy) for p in rotated_corners]

    def iou(self, other: Shape) -> Union[float, Literal[-1]]:
        """
        Calculate the Intersection over Union (IoU) with another shape.

        Args:
            other (Shape): The other shape to compare with.

        Returns:
            Union[float, Literal[-1]]: The IoU value or -1 if not applicable.
        """
        from datumaro.util.annotation_util import bbox_iou

        return bbox_iou(self.get_bbox(), other.get_bbox())

    def wrap(item, **kwargs):
        """
        Create a new RotatedBbox instance with updated attributes.

        Args:
            item (RotatedBbox): The original RotatedBbox instance.
            kwargs: Additional attributes to update.

        Returns:
            RotatedBbox: A new RotatedBbox instance with updated attributes.
        """
        d = {"x": item.x, "y": item.y, "w": item.w, "h": item.h, "r": item.r}
        d.update(kwargs)
        return attr.evolve(item, **d)


@attrs(slots=True, order=False)
class PointsCategories(Categories):
    """
    Describes (key-)point metainfo such as point names and joints.
    """

    @attrs(slots=True, order=False)
    class Category:
        # Names for specific points, e.g. eye, hose, mouth etc.
        # These labels are not required to be in LabelCategories
        labels: List[str] = field(factory=list, validator=default_if_none(list))

        # Pairs of connected point indices
        joints: Set[Tuple[int, int]] = field(factory=set, validator=default_if_none(set))

    items: Dict[int, Category] = field(factory=dict, validator=default_if_none(dict))

    @classmethod
    def from_iterable(
        cls,
        iterable: Union[
            Tuple[int, List[str]],
            Tuple[int, List[str], Set[Tuple[int, int]]],
        ],
    ) -> PointsCategories:
        """
        Create PointsCategories from an iterable.

        Args:
            iterable: An Iterable with the following elements:

                - a label id
                - a list of positional arguments for Categories

        Returns:
            PointsCategories: PointsCategories object
        """
        temp_categories = cls()

        for args in iterable:
            temp_categories.add(*args)
        return temp_categories

    def add(
        self,
        label_id: int,
        labels: Optional[Iterable[str]] = None,
        joints: Iterable[Tuple[int, int]] = None,
    ):
        if joints is None:
            joints = []
        joints = set(map(tuple, joints))
        self.items[label_id] = self.Category(labels, joints)

    def __contains__(self, idx: int) -> bool:
        return idx in self.items

    def __getitem__(self, idx: int) -> Category:
        return self.items[idx]

    def __len__(self) -> int:
        return len(self.items)


@attrs(slots=True, order=False)
class Points(Shape):
    """
    Represents an ordered set of points.

    Attributes:
        _type (AnnotationType): The type of annotation, set to `AnnotationType.points`.
        visibility (List[IntEnum]): A list indicating the visibility status of each point.

    Nested Class:
        Visibility (IntEnum): Enum representing the visibility state of points. It has three states:
            - absent: Point is absent (0).
            - hidden: Point is hidden (1).
            - visible: Point is visible (2).

    Methods:
        __attrs_post_init__: Validates that the number of points is even.
        get_area: Returns the area covered by the points, always zero.
        get_bbox: Returns the bounding box containing all visible or hidden points.
    """

    class Visibility(IntEnum):
        """
        Enum representing the visibility state of points.

        Attributes:
            absent (int): Point is absent (0).
            hidden (int): Point is hidden (1).
            visible (int): Point is visible (2).
        """

        absent = 0
        hidden = 1
        visible = 2

    _type = AnnotationType.points

    visibility: List[IntEnum] = field(default=None)

    @visibility.validator
    def _visibility_validator(self, attribute, visibility):
        """
        Validates and initializes the visibility list.

        Args:
            attribute: The attribute being validated.
            visibility (List[IntEnum]): A list indicating the visibility status of each point.

        Raises:
            AssertionError: If the length of the visibility list does not match half the length of the points list.
        """
        if visibility is None:
            visibility = [self.Visibility.visible] * (len(self.points) // 2)
        else:
            for i, v in enumerate(visibility):
                if not isinstance(v, self.Visibility):
                    visibility[i] = self.Visibility(v)
        assert len(visibility) == len(self.points) // 2
        self.visibility = visibility

    def __attrs_post_init__(self):
        """
        Validates that the number of points is even after initialization.

        Raises:
            AssertionError: If the number of points is not even.
        """
        assert len(self.points) % 2 == 0, self.points

    def get_area(self):
        """
        Returns the area covered by the points.

        Returns:
            int: Always returns 0.
        """
        return 0

    def get_bbox(self):
        """
        Returns the bounding box containing all visible or hidden points.

        Returns:
            List[float]: The bounding box as [x0, y0, width, height].
        """
        xs = [
            p
            for p, v in zip(self.points[0::2], self.visibility)
            if v != __class__.Visibility.absent
        ]
        ys = [
            p
            for p, v in zip(self.points[1::2], self.visibility)
            if v != __class__.Visibility.absent
        ]
        x0 = min(xs, default=0)
        x1 = max(xs, default=0)
        y0 = min(ys, default=0)
        y1 = max(ys, default=0)
        return [x0, y0, x1 - x0, y1 - y0]


@attrs(slots=True, order=False)
class Caption(Annotation):
    """
    Represents arbitrary text annotations.
    """

    _type = AnnotationType.caption
    caption: str = field(converter=str)


@attrs(slots=True, order=False, eq=False)
class _ImageAnnotation(Annotation):
    image: Image = field()

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        if not isinstance(other, __class__):
            return False
        return np.array_equal(self.image, other.image)


@attrs(slots=True, order=False, eq=False)
class SuperResolutionAnnotation(_ImageAnnotation):
    """
    Represents high resolution images.
    """

    _type = AnnotationType.super_resolution_annotation


@attrs(slots=True, order=False, eq=False)
class DepthAnnotation(_ImageAnnotation):
    """
    Represents depth images.
    """

    _type = AnnotationType.depth_annotation


@attrs(slots=True, init=False, order=False)
class Ellipse(Shape):
    """
    Ellipse represents an ellipse that is encapsulated by a rectangle.

    - x1 and y1 represent the top-left coordinate of the encapsulating rectangle
    - x2 and y2 representing the bottom-right coordinate of the encapsulating rectangle

    Parameters
    ----------

    x1: float
        left x coordinate of encapsulating rectangle
    y1: float
        top y coordinate of encapsulating rectangle
    x2: float
        right x coordinate of encapsulating rectangle
    y2: float
        bottom y coordinate of encapsulating rectangle
    """

    _type = AnnotationType.ellipse

    def __init__(self, x1: float, y1: float, x2: float, y2: float, *args, **kwargs):
        kwargs.pop("points", None)  # comes from wrap()
        self.__attrs_init__([x1, y1, x2, y2], *args, **kwargs)

    @property
    def x1(self):
        return self.points[0]

    @property
    def y1(self):
        return self.points[1]

    @property
    def x2(self):
        return self.points[2]

    @property
    def y2(self):
        return self.points[3]

    @property
    def w(self):
        return self.points[2] - self.points[0]

    @property
    def h(self):
        return self.points[3] - self.points[1]

    @property
    def c_x(self):
        return 0.5 * (self.points[0] + self.points[2])

    @property
    def c_y(self):
        return 0.5 * (self.points[1] + self.points[3])

    def get_area(self):
        return 0.25 * np.pi * self.w * self.h

    def get_bbox(self):
        return [self.x1, self.y1, self.w, self.h]

    def get_points(self, num_points: int = 720) -> List[Tuple[float, float]]:
        """
        Return points as a list of tuples, e.g. [(x0, y0), (x1, y1), ...].

        Parameters
        ----------
        num_points: int
            The number of boundary points of the ellipse.
            By default, one point is created for every 1 degree of interior angle (num_points=360).
        """
        points = self.as_polygon(num_points)

        return [(x, y) for x, y in zip(points[0::2], points[1::2])]

    def as_polygon(self, num_points: int = 720) -> List[float]:
        """
        Return a polygon as a list of tuples, e.g. [x0, y0, x1, y1, ...].

        Parameters
        ----------
        num_points: int
            The number of boundary points of the ellipse.
            By default, one point is created for every 1 degree of interior angle (num_points=360).
        """
        theta = np.linspace(0, 2 * np.pi, num=num_points)

        l1 = 0.5 * self.w
        l2 = 0.5 * self.h
        x_points = self.c_x + l1 * np.cos(theta)
        y_points = self.c_y + l2 * np.sin(theta)

        points = []
        for x, y in zip(x_points, y_points):
            points += [x, y]

        return points

    def iou(self, other: Shape) -> Union[float, Literal[-1]]:
        from datumaro.util.annotation_util import bbox_iou

        return bbox_iou(self.get_bbox(), other.get_bbox())

    def wrap(item: Ellipse, **kwargs) -> Ellipse:
        d = {"x1": item.x1, "y1": item.y1, "x2": item.x2, "y2": item.y2}
        d.update(kwargs)
        return attr.evolve(item, **d)


TableDtype = TypeVar("TableDtype", str, int, float)


@attrs(slots=True, order=False, eq=False)
class TabularCategories(Categories):
    """
    Describes tabular data metainfo such as column names and types.
    """

    @attrs(slots=True, order=False, eq=False)
    class Category:
        name: str = field(converter=str, validator=not_empty)
        dtype: Type[TableDtype] = field()
        labels: Set[Union[str, int]] = field(factory=set, validator=default_if_none(set))

        def __eq__(self, other):
            same_name = self.name == other.name
            same_dtype = self.dtype.__name__ == other.dtype.__name__
            same_labels = self.labels == other.labels
            return same_name and same_dtype and same_labels

        def __repr__(self):
            return f"name: {self.name}, dtype: {self.dtype.__name__}, labels: {self.labels}"

    items: List[Category] = field(factory=list, validator=default_if_none(list))
    _indices_by_name: Dict[str, int] = field(factory=dict, init=False, eq=False)

    @classmethod
    def from_iterable(
        cls,
        iterable: Iterable[
            Union[Tuple[str, Type[TableDtype]], Tuple[str, Type[TableDtype], Set[str]]]
        ],
    ) -> TabularCategories:
        """
        Creates a TabularCategories from iterable.

        Args:
            iterable: a list of (Category name, type) or (Category name, type, set of labels)

        Returns: a TabularCategories object
        """

        temp_categories = cls()

        for category in iterable:
            temp_categories.add(*category)

        return temp_categories

    def add(
        self,
        name: str,
        dtype: Type[TableDtype],
        labels: Optional[Set[str]] = None,
    ) -> int:
        """
        Add a Tabular Category.

        Args:
            name (str): Column name
            dtype (type): Type of the corresponding column. (str, int, or float)
            labels (optional, set(str)): Label values where the column can have.

        Returns:
            int: A index of added category.
        """
        assert name
        assert name not in self._indices_by_name
        assert dtype

        index = len(self.items)
        self.items.append(self.Category(name, dtype, labels))
        self._indices_by_name[name] = index

        return index

    def find(self, name: str) -> Tuple[Optional[int], Optional[Category]]:
        """
        Find Category information for the given column name.

        Args:
            name (str): Column name

        Returns:
            tuple(int, Category): A index and Category information.
        """
        index = self._indices_by_name.get(name)
        return index, self.items[index] if index is not None else None

    def __getitem__(self, index: int) -> Category:
        return self.items[index]

    def __contains__(self, name: str) -> bool:
        return self.find(name)[1] is not None

    def __len__(self) -> int:
        return len(self.items)

    def __iter__(self) -> Iterator[Category]:
        return iter(self.items)

    def __eq__(self, other) -> bool:
        if not super().__eq__(other):
            return False
        if not isinstance(other, __class__):
            return False
        return self.items == other.items


@attrs(slots=True, order=False)
class Tabular(Annotation):
    """
    Represents values of target columns in a tabular dataset.
    """

    _type = AnnotationType.tabular
    values: Dict[str, TableDtype] = field(converter=dict)


class Annotations(List[Annotation]):
    """List of `Annotation` equipped with additional utility functions."""

    def get_semantic_seg_mask(
        self, ignore_index: int = 0, dtype: np.dtype = np.uint8
    ) -> np.ndarray:
        """Extract semantic segmentation mask from a collection of Datumaro `Mask`s.

        Args:
            ignore_index: Scalar value to fill in the zeros in each binary mask
                before merging into a semantic segmentation mask. This value is usually used
                to represent a pixel denoting a not-interested region. Defaults to 0.
            dtype: Data type for the resulting mask. Defaults to np.uint8.

        Returns:
            Semantic segmentation mask generated by merging Datumaro `Mask`s.

        Raises:
            ValueError: If there are no mask annotations or if there is an inconsistency in mask sizes.
        """

        masks = [ann for ann in self if isinstance(ann, Mask)]
        # Mask with a lower z_order value will come first
        masks.sort(key=lambda mask: mask.z_order)

        if not masks:
            msg = "There is no mask annotations."
            raise ValueError(msg)

        # Dispatching for better performance
        # If all masks are `ExtractedMask`, share a same source `index_mask`, and
        # there is no label remapping.
        if (
            all(isinstance(mask, ExtractedMask) for mask in masks)
            # and set(id(mask.index_mask) for mask in masks) == 1
            and all(mask.index_mask == next(iter(masks)).index_mask for mask in masks)
            and all(mask.index == mask.label for mask in masks)
        ):
            index_mask = next(iter(masks)).index_mask
            semantic_seg_mask: np.ndarray = index_mask() if callable(index_mask) else index_mask
            if semantic_seg_mask.dtype != dtype:
                semantic_seg_mask = semantic_seg_mask.astype(dtype)

            labels = np.unique(np.array([mask.label for mask in masks]))
            ignore_index_mask = np.isin(semantic_seg_mask, labels, invert=True)

            return np.where(ignore_index_mask, ignore_index, semantic_seg_mask)

        class_masks = [mask.as_class_mask(ignore_index=ignore_index, dtype=dtype) for mask in masks]

        max_h = max([mask.shape[0] for mask in class_masks])
        max_w = max([mask.shape[1] for mask in class_masks])

        semantic_seg_mask = np.full(shape=(max_h, max_w), fill_value=ignore_index, dtype=dtype)

        for class_mask in class_masks:
            if class_mask.shape != semantic_seg_mask.shape:
                msg = f"There is inconsistency in mask size: {class_mask.shape}!={semantic_seg_mask.shape}."
                raise ValueError(msg, class_mask.shape, semantic_seg_mask.shape)

            ignore_index_mask = class_mask == ignore_index
            semantic_seg_mask = np.where(ignore_index_mask, semantic_seg_mask, class_mask)

        return semantic_seg_mask



# Copyright (C) 2019-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp
from collections import OrderedDict
from enum import Enum, auto
from itertools import chain
from typing import Dict, List, Optional, Tuple

import numpy as np

from datumaro.components.annotation import (
    AnnotationType,
    Colormap,
    LabelCategories,
    MaskCategories,
    RgbColor,
)
from datumaro.components.dataset_base import CategoriesInfo
from datumaro.components.errors import InvalidAnnotationError
from datumaro.util import dump_json_file, find, parse_json_file
from datumaro.util.meta_file_util import get_meta_file


class VocTask(Enum):
    voc = auto()
    voc_classification = auto()
    voc_detection = auto()
    voc_segmentation = auto()
    voc_instance_segmentation = auto()
    voc_action = auto()
    voc_layout = auto()


class VocLabel(Enum):
    background = 0
    aeroplane = 1
    bicycle = 2
    bird = 3
    boat = 4
    bottle = 5
    bus = 6
    car = 7
    cat = 8
    chair = 9
    cow = 10
    diningtable = 11
    dog = 12
    horse = 13
    motorbike = 14
    person = 15
    pottedplant = 16
    sheep = 17
    sofa = 18
    train = 19
    tvmonitor = 20
    ignored = 255


class VocPose(Enum):
    Unspecified = auto()
    Left = auto()
    Right = auto()
    Frontal = auto()
    Rear = auto()


class VocBodyPart(Enum):
    head = auto()
    hand = auto()
    foot = auto()


class VocAction(Enum):
    other = auto()
    jumping = auto()
    phoning = auto()
    playinginstrument = auto()
    reading = auto()
    ridingbike = auto()
    ridinghorse = auto()
    running = auto()
    takingphoto = auto()
    usingcomputer = auto()
    walking = auto()


def generate_colormap(length: int = 256) -> Colormap:
    def get_bit(number, index):
        return (number >> index) & 1

    colormap = np.zeros((length, 3), dtype=int)
    indices = np.arange(length, dtype=int)

    for j in range(7, -1, -1):
        for c in range(3):
            colormap[:, c] |= get_bit(indices, c) << j
        indices >>= 3

    return OrderedDict((id, tuple(color)) for id, color in enumerate(colormap))


VocColormap: Colormap = {
    id: color for id, color in generate_colormap(256).items() if id in {l.value for l in VocLabel}
}
VocInstColormap = generate_colormap(256)


class VocImporterType(Enum):
    default = auto()
    roboflow = auto()


class VocPath:
    IMAGES_DIR = "JPEGImages"
    ANNOTATIONS_DIR = "Annotations"
    SEGMENTATION_DIR = "SegmentationClass"
    INSTANCES_DIR = "SegmentationObject"
    SUBSETS_DIR = "ImageSets"
    IMAGE_EXT = ".jpg"
    SEGM_EXT = ".png"
    LABELMAP_FILE = "labelmap.txt"

    TASK_DIR = {
        VocTask.voc: "Main",
        VocTask.voc_classification: "Main",
        VocTask.voc_detection: "Main",
        VocTask.voc_segmentation: "Segmentation",
        VocTask.voc_instance_segmentation: "Segmentation",
        VocTask.voc_action: "Action",
        VocTask.voc_layout: "Layout",
    }


LabelMapConfig = Dict[str, Tuple[Optional[RgbColor], List[str], List[str]]]
"""A type representing a label map config"""
# Not totally type-correct, tuple elements are supposed to support modification.
# Therefore, the tuple is typically a list
# TODO: refactor, make type annotations conform with actual usage


def make_voc_label_map(task: VocTask = None) -> LabelMapConfig:
    if task == VocTask.voc_action:
        label_map = OrderedDict(
            {
                VocLabel.background.name: [VocColormap[VocLabel.background.value], [], []],
                VocLabel.person.name: [
                    VocColormap[VocLabel.person.value],
                    [],
                    [a.name for a in VocAction],
                ],
            }
        )
    elif task == VocTask.voc_layout:
        label_map = OrderedDict(
            {
                VocLabel.background.name: [VocColormap[VocLabel.background.value], [], []],
                VocLabel.person.name: [
                    VocColormap[VocLabel.person.value],
                    [p.name for p in VocBodyPart],
                    [],
                ],
            }
        )
    elif task in [
        VocTask.voc_classification,
        VocTask.voc_detection,
        VocTask.voc_segmentation,
        VocTask.voc_instance_segmentation,
    ]:
        labels = sorted(VocLabel, key=lambda l: l.value)
        label_map = OrderedDict(
            (label.name, [VocColormap[label.value], [], []]) for label in labels
        )
    else:
        labels = sorted(VocLabel, key=lambda l: l.value)
        label_map = OrderedDict(
            (label.name, [VocColormap[label.value], [], []]) for label in labels
        )
        label_map[VocLabel.person.name][1] = [p.name for p in VocBodyPart]
        label_map[VocLabel.person.name][2] = [a.name for a in VocAction]

    return label_map


def parse_label_map(path: str) -> LabelMapConfig:
    """
    Parses a label map file in the format:
    'name : color (r, g, b) : parts (hand, feet, ...) : actions (siting, standing, ...)'

    Parameters:
        path: File path

    Returns:
        A dictionary: label -> (color, parts, actions)
    """

    label_map = OrderedDict()
    with open(path, "r", encoding="utf-8") as f:
        for line in f:
            # skip empty and commented lines
            line = line.strip()
            if not line or line[0] == "#":
                continue

            # name : color : parts : actions
            label_desc = line.strip().split(":")
            if len(label_desc) != 4:
                raise InvalidAnnotationError(
                    f"Label description has wrong number of fields '{len(label_desc)}'. "
                    "Expected 4 ':'-separated fields."
                )
            name = label_desc[0]

            if name in label_map:
                raise InvalidAnnotationError(f"Label '{name}' is already defined in the label map")

            if 1 < len(label_desc) and label_desc[1]:
                color = label_desc[1].split(",")
                if len(color) != 3:
                    raise InvalidAnnotationError(
                        f"Label '{name}' has wrong color '{color}'. Expected an 'r,g,b' triplet."
                    )
                color = tuple(int(c) for c in color)
            else:
                color = None

            if 2 < len(label_desc) and label_desc[2]:
                parts = [s.strip() for s in label_desc[2].split(",")]
            else:
                parts = []

            if 3 < len(label_desc) and label_desc[3]:
                actions = [s.strip() for s in label_desc[3].split(",")]
            else:
                actions = []

            label_map[name] = [color, parts, actions]
    return label_map


def parse_meta_file(path: str) -> LabelMapConfig:
    # Uses custom format with extra fields
    meta_file = path
    if osp.isdir(path):
        meta_file = get_meta_file(path)

    dataset_meta = parse_json_file(meta_file)

    label_map = OrderedDict()
    parts = dataset_meta.get("parts", {})
    actions = dataset_meta.get("actions", {})

    for i, label in enumerate(dataset_meta.get("labels", [])):
        label_map[label] = [None, parts.get(str(i), []), actions.get(str(i), [])]

    colors = dataset_meta.get("segmentation_colors", [])

    for i, label in enumerate(dataset_meta.get("label_map", {}).values()):
        if label not in label_map:
            label_map[label] = [None, [], []]

        if any(colors) and colors[i] is not None:
            label_map[label][0] = tuple(colors[i])

    return label_map


def write_label_map(path: str, label_map: LabelMapConfig):
    with open(path, "w", encoding="utf-8") as f:
        f.write("# label:color_rgb:parts:actions\n")
        for label_name, label_desc in label_map.items():
            if label_desc[0]:
                color_rgb = ",".join(str(c) for c in label_desc[0])
            else:
                color_rgb = ""

            parts = ",".join(str(p) for p in label_desc[1])
            actions = ",".join(str(a) for a in label_desc[2])

            f.write("%s\n" % ":".join([label_name, color_rgb, parts, actions]))


def write_meta_file(path: str, label_map: LabelMapConfig):
    # Uses custom format with extra fields
    dataset_meta = {}

    labels = []
    labels_dict = {}
    segmentation_colors = []
    parts = {}
    actions = {}

    for i, (label_name, label_desc) in enumerate(label_map.items()):
        labels.append(label_name)
        if label_desc[0]:
            labels_dict[str(i)] = label_name
            segmentation_colors.append(
                [int(label_desc[0][0]), int(label_desc[0][1]), int(label_desc[0][2])]
            )

        parts[str(i)] = label_desc[1]
        actions[str(i)] = label_desc[2]

    dataset_meta["labels"] = labels

    if any(segmentation_colors):
        dataset_meta["label_map"] = labels_dict
        dataset_meta["segmentation_colors"] = segmentation_colors

        bg_label = find(label_map.items(), lambda x: x[1] == (0, 0, 0))
        if bg_label is not None:
            dataset_meta["background_label"] = str(bg_label[0])

    if any(parts):
        dataset_meta["parts"] = parts

    if any(actions):
        dataset_meta["actions"] = actions

    dump_json_file(get_meta_file(path), dataset_meta)


def make_voc_categories(
    label_map: Optional[LabelMapConfig] = None, task: Optional[VocTask] = VocTask.voc
) -> CategoriesInfo:
    if label_map is None:
        label_map = make_voc_label_map(task)

    categories = {}

    label_categories = LabelCategories()
    label_categories.attributes.update(["difficult", "truncated", "occluded"])

    for label, desc in label_map.items():
        label_categories.add(label, attributes=desc[2])

    if task in [VocTask.voc, VocTask.voc_layout]:
        for part in OrderedDict(
            (k, None) for k in chain(*(desc[1] for desc in label_map.values()))
        ):
            label_categories.add(part)

    categories[AnnotationType.label] = label_categories

    if task not in [VocTask.voc, VocTask.voc_segmentation, VocTask.voc_instance_segmentation]:
        return categories

    has_colors = any(v[0] is not None for v in label_map.values())
    if not has_colors:  # generate new colors
        colormap = generate_colormap(len(label_map))
    else:  # only copy defined colors
        label_id = lambda label: label_categories.find(label)[0]
        colormap = {
            label_id(name): desc[0] for name, desc in label_map.items() if desc[0] is not None
        }
    mask_categories = MaskCategories(colormap)
    mask_categories.inverse_colormap  # pylint: disable=pointless-statement
    categories[AnnotationType.mask] = mask_categories

    return categories



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import math
import time
from typing import Iterable, Optional, Tuple, TypeVar

from tqdm.auto import tqdm

T = TypeVar("T")


class ProgressReporter:
    """
    Only one set of methods must be called:
        - start - report_status - finish
        - iter
        - split

    This class is supposed to manage the state of children progress bars
    and release of their resources, if necessary.
    """

    @property
    def period(self) -> float:
        """
        Returns reporting period.

        For example, 0.1 would mean every 10%.
        """
        raise NotImplementedError

    @property
    def interval(self) -> float:
        """
        Returns reporting time interval in second.
        """
        raise NotImplementedError

    def start(self, total: int, *, desc: Optional[str] = None):
        """Initializes the progress bar"""
        raise NotImplementedError

    def report_status(self, progress: int):
        """Updates the progress bar"""
        raise NotImplementedError

    def finish(self):
        """Finishes the progress bar"""
        pass  # pylint: disable=unnecessary-pass

    def iter(
        self, iterable: Iterable[T], *, total: Optional[int] = None, desc: Optional[str] = None
    ) -> Iterable[T]:
        """
        Traverses the iterable and reports progress simultaneously.

        Starts and finishes the progress bar automatically.

        Args:
            iterable: An iterable to be traversed
            total: The expected number of iterations. If not provided, will
              try to use iterable.__len__.
            desc: The status message

        Returns:
            An iterable over elements of the input sequence
        """

        if total is None and hasattr(iterable, "__len__"):
            total = len(iterable)

        self.start(total, desc=desc)

        if total:
            display_step = math.ceil(total * self.period)

        s = time.time()
        for i, elem in enumerate(iterable):
            if (
                not total
                or (display_step and i % display_step == 0)
                or time.time() - s > self.interval
            ):
                s = time.time()
                self.report_status(i)

            yield elem

        self.finish()

    def split(self, count: int) -> Tuple[ProgressReporter, ...]:
        """
        Splits the progress bar into few independent parts.
        In case of 0 must return an empty tuple.

        This class is supposed to manage the state of children progress bars
        and release of their resources, if necessary.
        """
        raise NotImplementedError


class NullProgressReporter(ProgressReporter):
    @property
    def period(self) -> float:
        return 0

    @property
    def interval(self) -> float:
        return float("inf")

    def start(self, total: int, *, desc: Optional[str] = None):
        pass

    def report_status(self, progress: int):
        pass

    def iter(
        self, iterable: Iterable[T], *, total: Optional[int] = None, desc: Optional[str] = None
    ) -> Iterable[T]:
        yield from iterable

    def split(self, count: int) -> Tuple[ProgressReporter]:
        return (self,) * count


class SimpleProgressReporter(ProgressReporter):
    def __init__(self, period: float = 0.1, interval: float = float("inf")):
        self._period = period
        self._interval = interval

    @property
    def period(self) -> float:
        return self._period

    @property
    def interval(self) -> float:
        return self._interval

    def start(self, total: int, *, desc: Optional[str] = None):
        self._total = total
        self._desc = desc

    def report_status(self, progress: int):
        status = str(self._desc) if self._desc else ""
        status += (
            f" {progress:0{len(str(self._total))}d}/{self._total}"
            f" ({progress/self._total*100:6.2f}%)"
        )
        print(status)

    def finish(self):
        self.report_status(self._total)

    def split(self, count: int):
        return (SimpleProgressReporter(self._period, self._interval),) * count


class TQDMProgressReporter(ProgressReporter):
    def __init__(self, period: float = 0.1, interval: float = 0.1, **options):
        self._period = period
        self._interval = interval
        self._options = options

    @property
    def period(self) -> float:
        return self._period

    @property
    def interval(self) -> float:
        return self._interval

    def start(self, total: int, *, desc: Optional[str] = None):
        options = self._options.copy()
        if desc is not None:
            options["desc"] = desc
        self._total = total
        self._pbar = tqdm(total=total, **options)
        self._cur = 0

    def report_status(self, progress: int):
        self._pbar.update(progress - self._cur)
        self._cur = progress

    def finish(self):
        if self._total is None:
            self._total = self._cur  # Total can be None
        self._pbar.update(self._total - self._cur)
        self._pbar.close()

    def split(self, count: int):
        return (TQDMProgressReporter(self._period, self._interval, **self._options),) * count



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os
import re
from typing import List, Optional

from datumaro.components.importer import ImportContext, Importer
from datumaro.components.lazy_plugin import extra_deps
from datumaro.plugins.data_formats.tf_detection_api.base import TfDetectionApiBase
from datumaro.plugins.data_formats.tf_detection_api.format import TfrecordImporterType
from datumaro.util.tf_util import has_feature
from datumaro.util.tf_util import import_tf as _import_tf

tf = _import_tf()


@extra_deps("tensorflow")
class RoboflowTfrecordImporter(Importer):
    _ANNO_EXT = ".tfrecord"

    @classmethod
    def find_sources(cls, path):
        sources = cls._find_sources_recursive(
            path=path,
            ext=cls._ANNO_EXT,
            extractor_name="roboflow_tfrecord",
        )
        if len(sources) == 0:
            return []

        undesired_feature = {
            "image/source_id": tf.io.FixedLenFeature([], tf.string),
        }

        subsets = {}
        for source in sources:
            if has_feature(path=source["url"], feature=undesired_feature):
                continue
            subset_name = os.path.dirname(source["url"]).split(os.sep)[-1]
            subsets[subset_name] = source["url"]

        sources = [
            {
                "url": url,
                "format": "roboflow_tfrecord",
                "options": {
                    "subset": subset,
                },
            }
            for subset, url in subsets.items()
        ]

        return sources

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._ANNO_EXT]


class RoboflowTfrecordBase(TfDetectionApiBase):
    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        super().__init__(
            path=path,
            subset=subset,
            tfrecord_importer_type=TfrecordImporterType.roboflow,
            ctx=ctx,
        )

    @staticmethod
    def _parse_labelmap(text):
        entry_pattern = r'name:\s*"([^"]+)"\s*,\s*id:\s*(\d+)'
        entry_pattern = re.compile(entry_pattern)

        matches = re.findall(entry_pattern, text)

        labelmap = {name: int(id) for name, id in matches}

        return labelmap



# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import textwrap
from typing import Iterable, List


def add_subparser(subparsers, name, builder):
    return builder(lambda **kwargs: subparsers.add_parser(name, **kwargs))


class MultilineFormatter(argparse.HelpFormatter):
    """
    Keeps line breaks introduced with '|n' separator
    and spaces introduced with '|s',
    also removes the 'code-block' directives.
    """

    def __init__(self, keep_natural=False, **kwargs):
        super().__init__(**kwargs)
        self._keep_natural = keep_natural

    def _fill_text(self, text, width, indent):
        text = self._whitespace_matcher.sub(" ", text).strip()
        text = text.replace("|s", " ").replace(".. code-block::", "|n")

        paragraphs = text.split("|n ")
        if self._keep_natural:
            paragraphs = sum((p.split("\n ") for p in paragraphs), [])

        multiline_text = ""
        for paragraph in paragraphs:
            formatted_paragraph = (
                textwrap.fill(paragraph, width, initial_indent=indent, subsequent_indent=indent)
                + "\n"
            )
            multiline_text += formatted_paragraph
        return multiline_text

    def _format_action(self, action):
        """
        This requires to remove an ugly curly braced list generated by add_subparsers().
        Please see the following example.

        <w/o this patch>
            Commands:
            {get,describe}
                get           Download a publicly available dataset
                describe      Print information about downloadable datasets

        <w/ this patch>
            Commands:
                get           Download a publicly available dataset
                describe      Print information about downloadable datasets
        """
        parts = super()._format_action(action)
        if action.nargs == argparse.PARSER:
            parts = "\n".join(parts.split("\n")[1:])
        return parts


def required_count(nmin=0, nmax=0):
    assert 0 <= nmin and 0 <= nmax and nmin or nmax

    class RequiredCount(argparse.Action):
        def __call__(self, parser, args, values, option_string=None):
            k = len(values)
            if not ((nmin and (nmin <= k) or not nmin) and (nmax and (k <= nmax) or not nmax)):
                msg = "Argument '%s' requires" % self.dest
                if nmin and nmax:
                    msg += " from %s to %s arguments" % (nmin, nmax)
                elif nmin:
                    msg += " at least %s arguments" % nmin
                else:
                    msg += " no more %s arguments" % nmax
                raise argparse.ArgumentTypeError(msg)
            setattr(args, self.dest, values)

    return RequiredCount


def at_least(n):
    return required_count(n, 0)


def join_cli_args(args: argparse.Namespace, *names: Iterable[str]) -> List:
    "Merges arg values in a list"

    joined = []

    for name in names:
        value = getattr(args, name)
        if not isinstance(value, list):
            value = [value]
        joined += value

    return joined


def show_video_import_warning():
    log.warning(
        "Using 'video_frames' in a project may lead "
        "to different results across multiple runs, if the "
        "system setup changes (library version, OS, etc.). "
        "If you need stable results, consider splitting the video "
        "manually using instructions at: "
        "https://openvinotoolkit.github.io/datumaro/latest/docs/data-formats/media_formats"
    )


def make_subcommands_help(commands, help_line_start=0):
    desc = ""
    for command_name, command_type, command_help in commands:
        msg = ("  %-" + str(max(0, help_line_start - 2 - 1)) + "s%s\n") % (
            command_name,
            command_help,
        )
        if command_type is None:
            msg = msg.lstrip()  # un-indent for not subparser
        desc += msg

    return desc



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

DATUMARO_FORMAT_VERSION = "1.0"


class DatumaroPath:
    IMAGES_DIR = "images"
    ANNOTATIONS_DIR = "annotations"
    PCD_DIR = "point_clouds"
    VIDEO_DIR = "videos"
    MASKS_DIR = "masks"

    ANNOTATION_EXT = ".json"
    IMAGE_EXT = ".jpg"
    MASK_EXT = ".png"



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import List, Tuple

import cv2

from datumaro.components.abstracts import IModelInterpreter
from datumaro.components.abstracts.model_interpreter import LauncherInputType, ModelPred, PrepInfo
from datumaro.components.annotation import (
    Annotation,
    AnnotationType,
    FeatureVector,
    Label,
    LabelCategories,
)
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.errors import DatumaroError
from datumaro.components.media import Image


class GooglenetV4TfModelInterpreter(IModelInterpreter):
    LOGIT_KEY = "InceptionV4/Logits/Predictions"
    FEAT_KEY = "InceptionV4/Logits/PreLogitsFlatten/flatten_1/Reshape:0"

    def preprocess(self, inp: DatasetItem) -> Tuple[LauncherInputType, PrepInfo]:
        img = inp.media_as(Image).data
        img = cv2.resize(img, (299, 299))
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        return img, None

    def postprocess(self, pred: ModelPred, info: PrepInfo) -> List[Annotation]:
        logit = pred.get(self.LOGIT_KEY)
        if logit is None:
            raise DatumaroError(f'"{self.LOGIT_KEY}" key should exist in the model prediction.')

        feature_vector = pred.get(self.FEAT_KEY)
        if feature_vector is None:
            raise DatumaroError(f'"{self.FEAT_KEY}" key should exist in the model prediction.')

        outputs = [
            Label(label=label, attributes={"score": score}) for label, score in enumerate(logit)
        ]
        outputs += [FeatureVector(feature_vector)]

        return outputs  # [FeatureVector(logit), FeatureVector(feature_vector)]

    def get_categories(self):
        # output categories - label map etc.

        label_categories = LabelCategories()

        with open("samples/imagenet.class", "r", encoding="utf-8") as file:
            for line in file.readlines():
                label = line.strip()
                label_categories.add(label)

        return {AnnotationType.label: label_categories}



# Copyright (C) 2021-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import errno
import os
import os.path as osp
from typing import Dict, List, Optional, Tuple, Type, Union

import pandas as pd

from datumaro.components.annotation import AnnotationType, Categories, Tabular, TabularCategories
from datumaro.components.dataset_base import DatasetBase, DatasetItem
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Table, TableDtype, TableRow
from datumaro.util.os_util import find_files

# Only supports '.csv' extention.
TABULAR_EXTENSIONS = [
    "csv",
]


class TabularDataBase(DatasetBase):
    NAME = "tabular"

    def __init__(
        self,
        path: str,
        *,
        target: Optional[Union[str, List[str]]] = None,
        dtype: Optional[Dict[str, Type[TableDtype]]] = None,
        ctx: Optional[ImportContext] = None,
        **kwargs,
    ) -> None:
        """
        Read and compose a tabular dataset.
        The file name of each '.csv' file is regarded as subset.

        Args:
            path (str) : Path to a tabular dataset. (csv file or folder contains csv files).
            target (optional, str or list(str)) : Target column or list of target columns.
                If this is not specified (None), the last column is regarded as a target column.
                In case of a dataset with no targets, give an empty list as a parameter.
            dtype (optional, dict(str,str)) : Dictionay of column name -> type str ('str', 'int', or 'float').
                This can be used when automatic type inferencing is failed.
        """
        paths: List[str] = []
        if osp.isfile(path):
            paths.append(path)
        else:
            for path in find_files(path, TABULAR_EXTENSIONS):
                paths.append(path)

        if not paths:
            raise FileNotFoundError(errno.ENOENT, "Can't find tabular files", path)

        super().__init__(media_type=TableRow, ctx=ctx)

        self._infos = {"path": path}
        self._items, self._categories = self._parse(paths, target, dtype, **kwargs)

    def _parse(
        self,
        paths: List[str],
        target: Optional[Dict[str, List[str]]] = None,
        dtype: Optional[Dict[str, Type[TableDtype]]] = None,
        **kwargs,
    ) -> Tuple[List[DatasetItem], Dict[AnnotationType, Categories]]:
        """
        parse tabular files. Each file is regarded as a subset.

        Args:
            paths (list(str)) : A list of paths to tabular data files(csv files).
            target (optional, dict(str or list)) : Target column or list of target columns for each input and output.
                If this is not specified (None), the whole columns are regarded as a target column.
                In case of a dataset with no targets, give an empty list as a parameter.
            dtype (optional, dict(str,str)) : Dictionay of column name -> type str ('str', 'int', or 'float').
                This can be used when automatic type inferencing is failed.

        Returns:
            list (DatasetItem): dataset items
            dict (AnnotationType, Categories): categories info
        """
        # assert paths
        items: List[DatasetItem] = []
        categories: TabularCategories = TabularCategories()

        if target is not None:
            if "input" not in target or "output" not in target:
                raise TypeError('Target should have both "input" and "output"')

        for path in paths:
            table = Table.from_csv(path, dtype=dtype, **kwargs)

            targets: List[str] = []
            targets_ann: List[str] = []
            if target is None:
                targets.extend(table.columns)  # add all columns
            else:
                # add valid targeted output column name only
                if isinstance(target.get("input"), str) and target["input"] in table.columns:
                    targets.append(target["input"])
                elif isinstance(target.get("input"), list):
                    targets.extend(col for col in target["input"] if col in table.columns)
                if isinstance(target.get("output"), str) and target["output"] in table.columns:
                    targets_ann.append(target["output"])
                elif isinstance(target.get("output"), list):
                    targets_ann.extend(col for col in target["output"] if col in table.columns)
            targets = targets + targets_ann

            # set categories
            for target_ in targets_ann:
                _, category = categories.find(target_)
                target_dtype = table.dtype(target_)
                if target_dtype in [int, float, pd.api.types.CategoricalDtype()]:
                    # 'int' can be categorical, but we don't know this unless user gives information.
                    labels = set(
                        [
                            feature
                            for feature in table.features(target_, unique=True)
                            if not pd.isna(feature)
                        ]
                    )
                    if category is None:
                        categories.add(target_, target_dtype, labels)
                    else:  # update labels if they are different.
                        category.labels.union(labels)
                elif target_dtype is str:
                    if category is None:
                        categories.add(target_, target_dtype)
                else:
                    raise TypeError(
                        f"Unsupported type '{target_dtype}' for target column '{target_}'."
                    )

            # load annotations
            subset = osp.splitext(osp.basename(path))[0]
            row: TableRow
            table.select(targets)
            for row in table:  # type: TableRow
                id = f"{row.index}@{subset}"
                ann = [Tabular(values=row.data(targets_ann))] if targets_ann else None
                item = DatasetItem(
                    id=id,
                    subset=subset,
                    media=row,
                    annotations=ann,
                )
                items.append(item)

        return items, {AnnotationType.tabular: categories}

    def categories(self):
        return self._categories

    def __iter__(self):
        yield from self._items


def string_to_dict(input_string):
    pairs = input_string.split(",")
    result = {}

    for pair in pairs:
        split_pair = pair.split(":")
        # Check if the key is "input" or "output".
        if len(split_pair) == 2:
            key, value = split_pair
            if key == "input" or key == "output":
                if key in result:
                    result[key].append(value)
                else:
                    result[key] = [value]
            else:
                # Ignore other keys
                pass
        else:
            result[key].extend(split_pair)

    return result


class TabularDataImporter(Importer):
    """
    Import a tabular dataset.
    Each '.csv' file is regarded as a subset.
    """

    NAME = "tabular"

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--target",
            type=lambda x: string_to_dict(x),
            help="Target column or list of target columns for each input and output."
            "(ex. 'input:date,output:class', 'input:data,output:class,breed') (default:None)"
            "If this is not specified (None), the whole columns are regarded as a target column."
            "In case of a dataset with no targets, give an empty list as a parameter.",
        )
        parser.add_argument(
            "--dtype",
            type=lambda x: {k: v for k, v in (map.split(":") for map in x.split(","))},
            help="Type information for a column. (ex. 'date:str,x:int') (default:None) "
            "This can be used when automatic type inferencing is failed",
        )
        return parser

    @classmethod
    def find_sources(cls, path):
        if not osp.isdir(path):
            ext = osp.splitext(path)[1][1:]  # exclude "."
            if ext in TABULAR_EXTENSIONS:
                return [{"url": path, "format": TabularDataBase.NAME}]
        else:
            for _ in find_files(path, TABULAR_EXTENSIONS):  # find 1 depth only.
                return [{"url": path, "format": TabularDataBase.NAME}]
        return []

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return list({f".{ext}" for ext in TABULAR_EXTENSIONS})


class TabularDataExporter(Exporter):
    """
    Export a tabular dataset.
    This will save each subset into a '.csv' file regardless of 'save_media' value
    """

    NAME = "tabular"
    EXPORT_EXT = ".csv"
    DEFAULT_IMAGE_EXT = ".jpg"  # just to avoid assert error.

    def _apply_impl(self):
        extractor = self._extractor

        if extractor.media_type() and not issubclass(extractor.media_type(), TableRow):
            raise MediaTypeError("Media type is not a table.")

        # we don't check self._save_media.
        # regardless of the value, we always save media(csv) file.

        os.makedirs(self._save_dir, exist_ok=True)

        for sname in extractor.subsets():
            subset = extractor.get_subset(sname)
            path = osp.join(self._save_dir, sname + self.EXPORT_EXT)
            list_of_dicts: List[Dict[str, TableDtype]] = list()
            for item in subset:
                dicts = item.media.data()
                for ann in item.annotations:
                    if isinstance(ann, Tabular):
                        dicts.update(ann.values)  # update value
                list_of_dicts.append(dicts)

            table = Table.from_list(list_of_dicts)
            table.save(path)



# Copyright (C) 2020-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
import logging as log
import os
from collections import OrderedDict
from typing import Dict, Optional, Sequence, Set, Type

from datumaro.components.abstracts.merger import IMergerContext
from datumaro.components.annotation import AnnotationType
from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.dataset_base import CategoriesInfo, DatasetInfo, IDataset
from datumaro.components.dataset_item_storage import DatasetItemStorageDatasetView
from datumaro.components.errors import (
    ConflictingCategoriesError,
    DatasetMergeError,
    DatasetQualityError,
    MediaTypeError,
)
from datumaro.components.media import MediaElement
from datumaro.util import dump_json_file


class Merger(IMergerContext, CliPlugin):
    """Merge multiple datasets into one dataset"""

    def __init__(self, **options):
        super().__init__(**options)
        self.__dict__["_sources"] = None
        self.errors = []

    @staticmethod
    def merge_infos(sources: Sequence[DatasetInfo]) -> Dict:
        """Merge several :class:`IDataset` into one :class:`IDataset`"""
        infos = {}
        for source in sources:
            for k, v in source.items():
                if k in infos:
                    log.warning(
                        "Duplicated infos field %s: overwrite from %s to %s", k, infos[k], v
                    )
                infos[k] = v
        return infos

    @staticmethod
    def merge_categories(sources: Sequence[CategoriesInfo]) -> Dict:
        categories = {}
        for source_idx, source in enumerate(sources):
            for cat_type, source_cat in source.items():
                existing_cat = categories.setdefault(cat_type, source_cat)
                if existing_cat != source_cat and len(source_cat) != 0:
                    if len(existing_cat) == 0:
                        categories[cat_type] = source_cat
                    else:
                        raise ConflictingCategoriesError(
                            "Merging of datasets with different categories is "
                            "only allowed in 'merge' command.",
                            sources=list(range(source_idx)),
                        )
        return categories

    @staticmethod
    def merge_media_types(sources: Sequence[IDataset]) -> Optional[Type[MediaElement]]:
        if sources:
            media_type = sources[0].media_type()
            for s in sources:
                if not issubclass(s.media_type(), media_type) or not issubclass(
                    media_type, s.media_type()
                ):
                    # Symmetric comparision is needed in the case of subclasses:
                    # eg. Image and RoIImage
                    raise MediaTypeError("Datasets have different media types")
            return media_type

        return None

    @staticmethod
    def merge_ann_types(sources: Sequence[IDataset]) -> Optional[Set[AnnotationType]]:
        ann_types = set()
        for source in sources:
            ann_types.union(source.ann_types())
        return ann_types

    def __call__(self, *datasets: IDataset) -> DatasetItemStorageDatasetView:
        infos = self.merge_infos(d.infos() for d in datasets)
        categories = self.merge_categories(d.categories() for d in datasets)
        media_type = self.merge_media_types(datasets)
        ann_types = self.merge_ann_types(datasets)
        return DatasetItemStorageDatasetView(
            parent=self.merge(datasets),
            infos=infos,
            categories=categories,
            media_type=media_type,
            ann_types=ann_types,
        )

    def save_merge_report(self, path: str) -> None:
        item_errors = OrderedDict()
        source_errors = OrderedDict()
        all_errors = []

        for e in self.errors:
            if isinstance(e, DatasetQualityError):
                item_errors[str(e.item_id)] = item_errors.get(str(e.item_id), 0) + 1
            elif isinstance(e, DatasetMergeError):
                for s in e.sources:
                    source_errors[str(s)] = source_errors.get(s, 0) + 1
                item_errors[str(e.item_id)] = item_errors.get(str(e.item_id), 0) + 1

            all_errors.append(str(e))

        errors = OrderedDict(
            [
                ("Item errors", item_errors),
                ("Source errors", source_errors),
                ("All errors", all_errors),
            ]
        )

        os.makedirs(os.path.dirname(path), exist_ok=True)

        dump_json_file(path, errors, indent=True)

    def get_any_label_name(self, ann, label_id):
        raise NotImplementedError



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import List, Optional, Set

from datumaro.components.abstracts.merger import IMatcherContext
from datumaro.components.annotation import Annotation, AnnotationType, LabelCategories
from datumaro.components.annotations.matcher import BboxMatcher, match_segments_more_than_pair
from datumaro.components.dataset_base import DatasetItem, IDataset
from datumaro.components.launcher import Launcher
from datumaro.components.transformer import ModelTransform


class MissingAnnotationDetection(ModelTransform):
    """This class is used to find annotations that are missing from the ground truth annotations.

    To accomplish this, it generates model predictions for the dataset using the given launcher.
    However, not all of these model predictions can be considered missing annotations since the dataset
    already contains ground truth annotations and some of the model predictions can be duplicated with them.
    To identify the missing annotations from the model predictions, this class filters out the predictions
    that spatially overlap with the ground truth annotations.

    For example, the follwing example will produce ``[Bbox(1, 1, 1, 1, label=1, attributes={"score": 0.5})]`` as
    the missing annotations since ``Bbox(0, 0, 1, 1, label=1, attributes={"score": 1.0})`` is overlapped with the
    ground-truth annotation, ``Bbox(0, 0, 1, 1, label=0)`` (``label_agnostic_matching=True``)

    .. code-block:: python

        ground_truth_annotations = [
            Bbox(0, 0, 1, 1, label=0),
            Bbox(1, 0, 1, 1, label=1),
            Bbox(0, 1, 1, 1, label=2),
        ]
        model_predictions = [
            Bbox(0, 0, 1, 1, label=1, attributes={"score": 1.0}),
            Bbox(1, 1, 1, 1, label=1, attributes={"score": 0.5}),
        ]

    Args:
        extractor: The dataset used to find missing labeled annotations.
        launcher: The launcher used to generate model predictions from the dataset.
        batch_size: The size of the batches used during processing.
        pairwise_dist: The distance metric used to measure the distance between two annotations.
            Typically, the distance metric is Intersection over Union (IoU), which is bounded between 0 and 1.
        score_threshold: The minimum score required for an annotation to be considered
            a candidate for missing annotations.
        label_agnostic_matching: If set to false, annotations with different labels are not matched
            to determine their spatial overlap. In the above example, ``label_agnostic_matching=False``
            will produce ``model_predictions`` as is since ``Bbox(0, 0, 1, 1, label=1, attributes={"score": 1.0})``
            has different label with ``Bbox(0, 0, 1, 1, label=0)``.
    """

    def __init__(
        self,
        extractor: IDataset,
        launcher: Launcher,
        batch_size: int = 1,
        pairwise_dist: float = 0.75,
        score_threshold: Optional[float] = None,
        label_agnostic_matching: bool = True,
    ):
        super().__init__(extractor, launcher, batch_size, append_annotation=False)
        self._score_threshold = score_threshold

        class LabelAgnosticMatcherContext(IMatcherContext):
            def get_any_label_name(self, ann: Annotation, label_id: int) -> str:
                return ""

        label_categories: LabelCategories = self.categories()[AnnotationType.label]

        class LabelSpecificMatcherContext(IMatcherContext):
            def get_any_label_name(self, ann: Annotation, label_id: int) -> str:
                return label_categories[label_id]

        self._support_matchers = {
            AnnotationType.bbox: BboxMatcher(
                pairwise_dist=pairwise_dist,
                context=LabelAgnosticMatcherContext()
                if label_agnostic_matching
                else LabelSpecificMatcherContext(),
                match_segments=match_segments_more_than_pair,
            ),
        }

    def _process_batch(
        self,
        batch: List[DatasetItem],
    ) -> List[DatasetItem]:
        inference = self._launcher.launch(
            batch=[item for item in batch if self._launcher.type_check(item)]
        )

        for annotations in inference:
            self._check_annotations(annotations)

        return [
            self.wrap_item(
                item,
                annotations=self._find_missing_anns(
                    gt_anns=item.annotations,
                    pseudo_anns=self._apply_score_threshold(annotations),
                ),
            )
            for item, annotations in zip(batch, inference)
        ]

    def _apply_score_threshold(self, annotations: List[Annotation]) -> List[Annotation]:
        if self._score_threshold is None:
            return annotations

        return [
            ann for ann in annotations if ann.attributes.get("score", 1.0) > self._score_threshold
        ]

    def _find_missing_anns(
        self, gt_anns: List[Annotation], pseudo_anns: List[Annotation]
    ) -> List[Annotation]:
        ids_of_pseudo_anns = set(id(ann) for ann in pseudo_anns)

        missing_labeled_anns = []
        for ann_type, matcher in self._support_matchers.items():
            clusters = matcher.match_annotations(
                [
                    [ann for ann in gt_anns if ann.type == ann_type],
                    [ann for ann in pseudo_anns if ann.type == ann_type],
                ]
            )
            for cluster in clusters:
                ann = self._pick_missing_ann_from_cluster(cluster, ids_of_pseudo_anns)
                if ann is not None:
                    missing_labeled_anns.append(ann)

        return missing_labeled_anns

    @staticmethod
    def _pick_missing_ann_from_cluster(
        cluster: List[Annotation], ids_of_pseudo_anns: Set[int]
    ) -> Optional[Annotation]:
        assert len(cluster) > 0, "cluster should not be empty."
        pseudo_label_anns = []
        gt_label_anns = []

        for ann in cluster:
            if id(ann) in ids_of_pseudo_anns:
                pseudo_label_anns.append(ann)
            else:
                gt_label_anns.append(ann)

        if len(gt_label_anns) > 0:
            return None

        return max(pseudo_label_anns, key=lambda ann: ann.attributes.get("score", -1))



# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import itertools
import logging as log
import os.path as osp
from importlib.util import find_spec
from types import SimpleNamespace as namespace
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

import attrs
import numpy as np
from attrs import field, frozen

from datumaro.components.annotation import AnnotationType, Bbox, Label, LabelCategories
from datumaro.components.dataset_base import CategoriesInfo, DatasetInfo, DatasetItem, IDataset
from datumaro.components.media import Image, MediaElement
from datumaro.util.import_util import lazy_import
from datumaro.util.tf_util import import_tf

TFDS_EXTRACTOR_AVAILABLE = True if find_spec("tensorflow_datasets") is not None else False

if TFDS_EXTRACTOR_AVAILABLE:
    try:
        if TYPE_CHECKING:
            tf = import_tf()
            import tensorflow_datasets as tfds
        else:
            tfds = lazy_import("tensorflow_datasets")
    except ImportError:
        log.debug(
            "Unable to import TensorFlow or TensorFlow Datasets. "
            "Dataset downloading via TFDS is disabled."
        )


@frozen(kw_only=True)
class TfdsDatasetMetadata:
    # If you add attributes to this class, make sure to update the reporting logic
    # in the `describe-downloads` command to include them.

    human_name: str
    default_output_format: str
    media_type: Type[MediaElement]

    # Every TFDS dataset has a home page (the TFDS documentation page), but
    # this field is still optional for a couple of reasons:
    # * We might want to reuse this class later if we implement downloading
    #   without the use of TFDS, and at that point it might happen that a dataset
    #   has no home page (e.g. it's been taken down).
    # * It's convenient to initially leave this field blank and mass-update it later,
    #   and in order for us to not break typing, it should be optional.
    home_url: Optional[str] = None


@frozen
class _TfdsAdapter:
    category_transformers: Sequence[
        Callable[[tfds.core.DatasetBuilder, CategoriesInfo, namespace], None]
    ]
    data_transformers: Sequence[Callable[[Any, DatasetItem, namespace], None]]
    id_generator: Callable[[Any], str] = field(default=None, kw_only=True)

    metadata: TfdsDatasetMetadata

    def transform_categories(
        self,
        tfds_builder: tfds.core.DatasetBuilder,
        categories: CategoriesInfo,
        state: namespace,
    ) -> None:
        for t in self.category_transformers:
            t(tfds_builder, categories, state)

    def transform_data(
        self,
        tfds_example: Any,
        item: DatasetItem,
        state: namespace,
    ) -> None:
        for t in self.data_transformers:
            t(tfds_example, item, state)


_FeaturePath = Union[str, Tuple[str, ...]]


def _resolve_feature_path(
    feature_path: _FeaturePath,
    root: tfds.features.FeaturesDict,
) -> tfds.features.FeatureConnector:
    if isinstance(feature_path, str):
        return root[feature_path]

    feature_connector = root

    for segment in feature_path:
        assert isinstance(
            feature_connector,
            (
                tfds.features.FeaturesDict,
                tfds.features.Sequence,
            ),
        )

        if isinstance(feature_connector, tfds.features.Sequence):
            assert segment == "feature"
            feature_connector = feature_connector.feature
        else:
            feature_connector = feature_connector[segment]

    return feature_connector


@frozen
class _SetLabelCategoriesFromClassLabelFeature:
    feature_path: _FeaturePath

    def __call__(
        self,
        tfds_builder: tfds.core.DatasetBuilder,
        categories: CategoriesInfo,
        state: namespace,
    ) -> None:
        assert AnnotationType.label not in categories

        feature_connector = _resolve_feature_path(self.feature_path, tfds_builder.info.features)

        assert isinstance(feature_connector, tfds.features.ClassLabel)
        categories[AnnotationType.label] = LabelCategories.from_iterable(feature_connector.names)


@frozen
class _SetImageFromImageFeature:
    feature_name: str
    filename_feature_name: Optional[str] = field(default=None)

    def __call__(
        self,
        tfds_example: Any,
        item: DatasetItem,
        state: namespace,
    ) -> None:
        if self.filename_feature_name:
            filename = tfds_example[self.filename_feature_name].numpy().decode("UTF-8")
            if osp.exists(filename):
                item.media = Image.from_file(path=filename)
                return
        data = tfds_example[self.feature_name].numpy()
        if isinstance(data, np.ndarray):
            item.media = Image.from_numpy(data=data)
        else:
            item.media = Image.from_bytes(data=data)


@frozen
class _AddLabelFromClassLabelFeature:
    feature_name: str

    def __call__(
        self,
        tfds_example: Any,
        item: DatasetItem,
        state: namespace,
    ) -> None:
        item.annotations.append(
            Label(tfds_example[self.feature_name].numpy()),
        )


@frozen
class _AttributeMemberMapping:
    member_name: str
    attribute_name: str = field()
    value_converter: Optional[Callable[[Any, namespace], Any]] = None

    @attribute_name.default
    def _attribute_name_default(self):
        return self.member_name


@frozen
class _AddObjectsFromFeature:
    feature_name: str
    bbox_member: str
    label_member: Optional[str] = field(default=None, kw_only=True)
    attribute_members: Tuple[_AttributeMemberMapping, ...] = field(
        default=(),
        kw_only=True,
        converter=lambda values: tuple(
            value if isinstance(value, _AttributeMemberMapping) else _AttributeMemberMapping(value)
            for value in values
        ),
    )

    def __call__(
        self,
        tfds_example: Any,
        item: DatasetItem,
        state: namespace,
    ) -> None:
        tfds_objects = tfds_example[self.feature_name]
        tfds_bboxes = tfds_objects[self.bbox_member]
        num_objects = tfds_bboxes.shape[0]

        tfds_labels = None
        if self.label_member is not None:
            tfds_labels = tfds_objects[self.label_member]
            assert tfds_labels.shape[0] == num_objects

        for am_mapping in self.attribute_members:
            assert tfds_objects[am_mapping.member_name].shape[0] == num_objects

        for i in range(num_objects):
            norm_ymin, norm_xmin, norm_ymax, norm_xmax = tfds_bboxes[i].numpy()

            new_bbox = Bbox(
                x=norm_xmin * item.media.size[1],
                y=norm_ymin * item.media.size[0],
                w=(norm_xmax - norm_xmin) * item.media.size[1],
                h=(norm_ymax - norm_ymin) * item.media.size[0],
            )

            if tfds_labels is not None:
                new_bbox.label = tfds_labels[i].numpy()

            for am_mapping in self.attribute_members:
                attr_value = tfds_objects[am_mapping.member_name][i].numpy()

                if am_mapping.value_converter:
                    attr_value = am_mapping.value_converter(attr_value, state)

                new_bbox.attributes[am_mapping.attribute_name] = attr_value

            item.annotations.append(new_bbox)


@frozen
class _SetAttributeFromFeature:
    feature_name: str
    attribute_name: str

    def __call__(
        self,
        tfds_example: Any,
        item: DatasetItem,
        state: namespace,
    ) -> None:
        item.attributes[self.attribute_name] = tfds_example[self.feature_name].numpy()


@frozen
class _GenerateIdFromTextFeature:
    feature_name: str

    def __call__(self, tfds_example: Any) -> str:
        return tfds_example[self.feature_name].numpy().decode("UTF-8")


@frozen
class _GenerateIdFromFilenameFeature:
    feature_name: str

    def __call__(self, tfds_example: Any) -> str:
        file_name = tfds_example[self.feature_name].numpy().decode("UTF-8")
        return osp.splitext(file_name)[0]


_MNIST_ADAPTER = _TfdsAdapter(
    category_transformers=[_SetLabelCategoriesFromClassLabelFeature("label")],
    data_transformers=[
        _SetImageFromImageFeature("image"),
        _AddLabelFromClassLabelFeature("label"),
    ],
    metadata=TfdsDatasetMetadata(
        human_name="MNIST",
        default_output_format="mnist",
        media_type=Image,
    ),
)

_CIFAR_ADAPTER = _TfdsAdapter(
    category_transformers=[_SetLabelCategoriesFromClassLabelFeature("label")],
    data_transformers=[
        _SetImageFromImageFeature("image"),
        _AddLabelFromClassLabelFeature("label"),
    ],
    id_generator=_GenerateIdFromTextFeature("id"),
    metadata=TfdsDatasetMetadata(
        human_name="CIFAR", default_output_format="cifar", media_type=Image
    ),
)

_COCO_ADAPTER = _TfdsAdapter(
    category_transformers=[
        _SetLabelCategoriesFromClassLabelFeature(("objects", "feature", "label"))
    ],
    data_transformers=[
        _SetImageFromImageFeature("image", filename_feature_name="image/filename"),
        _AddObjectsFromFeature(
            "objects", "bbox", label_member="label", attribute_members=("is_crowd",)
        ),
        _SetAttributeFromFeature("image/id", "id"),
    ],
    id_generator=_GenerateIdFromFilenameFeature("image/filename"),
    metadata=TfdsDatasetMetadata(
        human_name="COCO", default_output_format="coco_instances", media_type=Image
    ),
)

_IMAGENET_ADAPTER = _TfdsAdapter(
    category_transformers=[_SetLabelCategoriesFromClassLabelFeature("label")],
    data_transformers=[
        _SetImageFromImageFeature("image", filename_feature_name="file_name"),
        _AddLabelFromClassLabelFeature("label"),
    ],
    id_generator=_GenerateIdFromFilenameFeature("file_name"),
    metadata=TfdsDatasetMetadata(
        human_name="ImageNet", default_output_format="imagenet_txt", media_type=Image
    ),
)

_EUROSAT_ADAPTER = _TfdsAdapter(
    category_transformers=[_SetLabelCategoriesFromClassLabelFeature("label")],
    data_transformers=[
        _SetImageFromImageFeature("image"),
        _AddLabelFromClassLabelFeature("label"),
    ],
    id_generator=_GenerateIdFromFilenameFeature("filename"),
    metadata=TfdsDatasetMetadata(
        human_name="EuroSAT", default_output_format="imagenet_txt", media_type=Image
    ),
)


_UC_MERCED_ADAPTER = _TfdsAdapter(
    category_transformers=[_SetLabelCategoriesFromClassLabelFeature("label")],
    data_transformers=[
        _SetImageFromImageFeature("image"),
        _AddLabelFromClassLabelFeature("label"),
    ],
    id_generator=_GenerateIdFromFilenameFeature("filename"),
    metadata=TfdsDatasetMetadata(
        human_name="UCMerced", default_output_format="imagenet_txt", media_type=Image
    ),
)


def _voc_save_pose_names(
    tfds_builder: tfds.core.DatasetBuilder,
    categories: CategoriesInfo,
    state: namespace,
) -> None:
    # TFDS represents poses as indexes, but Datumaro represents them as strings.
    # To convert between representations, save the pose names at the start and
    # use them when we're converting boxes.
    # TFDS also provides the pose names in lower case, even though they're title
    # case in the original dataset. Fix them back to title case so that the
    # output dataset better resembles the original dataset.

    state.pose_names = [
        name.title() for name in tfds_builder.info.features["objects"].feature["pose"].names
    ]


_VOC_ADAPTER = _TfdsAdapter(
    category_transformers=[
        _SetLabelCategoriesFromClassLabelFeature(("objects", "feature", "label")),
        _voc_save_pose_names,
    ],
    data_transformers=[
        _SetImageFromImageFeature("image", filename_feature_name="image/filename"),
        _AddObjectsFromFeature(
            "objects",
            "bbox",
            label_member="label",
            attribute_members=(
                _AttributeMemberMapping("is_difficult", "difficult"),
                _AttributeMemberMapping("is_truncated", "truncated"),
                _AttributeMemberMapping(
                    "pose", value_converter=lambda idx, state: state.pose_names[idx]
                ),
            ),
        ),
    ],
    id_generator=_GenerateIdFromFilenameFeature("image/filename"),
    metadata=TfdsDatasetMetadata(
        human_name="PASCAL VOC", default_output_format="voc", media_type=Image
    ),
)


def _evolve_adapter_meta(adapter: _TfdsAdapter, **kwargs):
    return attrs.evolve(adapter, metadata=attrs.evolve(adapter.metadata, **kwargs))


_TFDS_ADAPTERS = {
    "cifar10": _evolve_adapter_meta(_CIFAR_ADAPTER, human_name="CIFAR-10"),
    "cifar100": _evolve_adapter_meta(_CIFAR_ADAPTER, human_name="CIFAR-100"),
    "coco/2014": _evolve_adapter_meta(_COCO_ADAPTER, human_name="COCO (2014-2015)"),
    "imagenet_v2": _evolve_adapter_meta(_IMAGENET_ADAPTER, human_name="ImageNetV2"),
    "mnist": _MNIST_ADAPTER,
    "voc/2012": _evolve_adapter_meta(_VOC_ADAPTER, human_name="PASCAL VOC 2012"),
    "eurosat": _evolve_adapter_meta(_EUROSAT_ADAPTER, human_name="EuroSAT"),
    "uc_merced": _evolve_adapter_meta(_UC_MERCED_ADAPTER, human_name="UCMerced"),
}

# Assign the TFDS catalog page as the documentation URL for all datasets.
_TFDS_ADAPTERS = {
    name: _evolve_adapter_meta(
        adapter,
        home_url="https://www.tensorflow.org/datasets/catalog/" + name.split("/", maxsplit=1)[0],
    )
    for name, adapter in _TFDS_ADAPTERS.items()
}


class _TfdsSplitExtractor(IDataset):
    def __init__(
        self,
        parent: _TfdsExtractor,
        tfds_split: tf.data.Dataset,
        tfds_split_info: tfds.core.SplitInfo,
    ):
        self._parent = parent
        self._tfds_split = tfds_split
        self._tfds_split_info = tfds_split_info

    def __len__(self) -> int:
        return self._tfds_split_info.num_examples

    def __iter__(self) -> Iterator[DatasetItem]:
        for example_index, tfds_example in enumerate(self._tfds_split):
            if self._parent._adapter.id_generator:
                item_id = self._parent._adapter.id_generator(tfds_example)
            else:
                item_id = str(example_index)

            dm_item = DatasetItem(id=item_id, subset=self._tfds_split_info.name)
            self._parent._adapter.transform_data(tfds_example, dm_item, self._parent._state)

            yield dm_item

    def infos(self) -> DatasetInfo:
        return self._parent.infos()

    def categories(self) -> CategoriesInfo:
        return self._parent.categories()

    def subsets(self) -> Dict[str, IDataset]:
        return {self._tfds_split_info.name: self}

    def get_subset(self, name) -> IDataset:
        assert name == self._tfds_split_info.name
        return self

    def get(self, id, subset=None) -> Optional[DatasetItem]:
        if subset is not None and subset != self._tfds_split_info.name:
            return None

        for item in self:
            if item.id == id:
                return item

        return None

    def media_type(self) -> Type[MediaElement]:
        return self._parent._media_type

    def ann_types(self) -> Set[AnnotationType]:
        return self._parent.ann_types()


class _TfdsExtractor(IDataset):
    _categories: CategoriesInfo
    _infos: DatasetInfo

    def __init__(self, tfds_ds_name: str) -> None:
        self._adapter = _TFDS_ADAPTERS[tfds_ds_name]
        tfds_builder = tfds.builder(tfds_ds_name)
        tfds_ds_info = tfds_builder.info

        self._infos = {}
        self._categories = {}
        self._state = namespace()
        self._adapter.transform_categories(tfds_builder, self._categories, self._state)
        self._media_type = self._adapter.metadata.media_type

        tfds_decoders = {}
        for tfds_feature_name, tfds_fc in tfds_ds_info.features.items():
            if isinstance(tfds_fc, tfds.features.Image):
                tfds_decoders[tfds_feature_name] = tfds.decode.SkipDecoding()

        tfds_builder.download_and_prepare()
        self._tfds_ds = tfds_builder.as_dataset(decoders=tfds_decoders)

        self._split_extractors = {
            split_name: _TfdsSplitExtractor(self, split, tfds_ds_info.splits[split_name])
            # Since dicts in Python 3.7+ (and de facto in 3.6+) are
            # order-preserving, sort the splits by name so that we always
            # iterate over them in alphabetical order.
            for split_name, split in sorted(self._tfds_ds.items())
        }

    def __len__(self) -> int:
        return sum(len(ex) for ex in self._split_extractors.values())

    def __iter__(self) -> Iterator[DatasetItem]:
        return itertools.chain.from_iterable(self._split_extractors.values())

    def infos(self) -> DatasetInfo:
        return self._infos

    def categories(self) -> CategoriesInfo:
        return self._categories

    def subsets(self) -> Dict[str, IDataset]:
        return self._split_extractors

    def get_subset(self, name) -> IDataset:
        return self._split_extractors[name]

    def get(self, id, subset=None) -> Optional[DatasetItem]:
        if subset is None:
            for ex in self._split_extractors.values():
                item = ex.get(id)
                if item is not None:
                    return item
            return None

        if subset not in self._split_extractors:
            return None
        return self._split_extractors[subset].get(id)

    def media_type(self) -> Type[MediaElement]:
        return self._media_type

    def ann_types(self) -> Set[AnnotationType]:
        ann_types = set()
        for items in self._split_extractors.values():
            for item in items:
                for ann in item.annotations:
                    ann_types.add(ann.type)
        return ann_types


# Some dataset metadata elements are either inconvenient to hardcode, or may change
# depending on the version of TFDS. We fetch them from the attributes of the `tfds.Builder`
# object. However, creating the builder may be time-consuming, because if the dataset
# is not already downloaded, TFDS fetches some data from its Google Cloud bucket.
# We therefore only fetch this metadata (which we call _remote_ metadata) when
# we actually need it.

# If you add attributes to either of these metadata classes, make sure to update
# the reporting logic in the `describe-downloads` command to include them.


@frozen(kw_only=True)
class TfdsSubsetRemoteMetadata:
    num_items: int


@frozen(kw_only=True)
class TfdsDatasetRemoteMetadata(TfdsDatasetMetadata):
    # For convenience, the remote metadata also includes the local metadata.
    description: str
    download_size: int
    subsets: Mapping[str, TfdsSubsetRemoteMetadata]

    # the dataset might not have the notion of classes, so `num_classes` may
    # be None.
    num_classes: Optional[int]
    version: str


class TfdsDataset:
    def __init__(self, tfds_ds_name: str):
        self._tfds_ds_name = tfds_ds_name
        self._adapter = _TFDS_ADAPTERS[tfds_ds_name]

    @property
    def metadata(self) -> TfdsDatasetMetadata:
        return self._adapter.metadata

    def make_extractor(self) -> IDataset:
        return _TfdsExtractor(self._tfds_ds_name)

    def query_remote_metadata(self) -> TfdsDatasetRemoteMetadata:
        tfds_builder = tfds.builder(self._tfds_ds_name)

        categories = {}
        state = namespace()
        self._adapter.transform_categories(tfds_builder, categories, state)

        num_classes = None
        if AnnotationType.label in categories:
            num_classes = len(categories[AnnotationType.label])

        return TfdsDatasetRemoteMetadata(
            **attrs.asdict(self._adapter.metadata),
            description=tfds_builder.info.description,
            download_size=int(tfds_builder.info.download_size),
            num_classes=num_classes,
            subsets={
                name: TfdsSubsetRemoteMetadata(num_items=split_info.num_examples)
                for name, split_info in tfds_builder.info.splits.items()
            },
            version=str(tfds_builder.info.version),
        )


AVAILABLE_TFDS_DATASETS: Mapping[str, TfdsDataset] = {
    name: TfdsDataset(name) for name in _TFDS_ADAPTERS
}



# Copyright (C) 2020-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging
from contextlib import contextmanager
from io import StringIO


@contextmanager
def logging_disabled(max_level=logging.CRITICAL):
    previous_level = logging.root.manager.disable
    logging.disable(max_level)
    try:
        yield
    finally:
        logging.disable(previous_level)


@contextmanager
def catch_logs(logger=None):
    logger = logging.getLogger(logger)

    old_propagate = logger.propagate
    prev_handlers = logger.handlers

    stream = StringIO()
    handler = logging.StreamHandler(stream)
    logger.handlers = [handler]
    logger.propagate = False

    try:
        yield stream
    finally:
        logger.handlers = prev_handlers
        logger.propagate = old_propagate



# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

import logging as log
import math
import random
import warnings
from collections import defaultdict
from typing import TYPE_CHECKING, Iterable, List, Optional, Tuple, Union, overload

import cv2
import numpy as np
from PIL import ImageColor

from datumaro.components.annotation import (
    Annotation,
    AnnotationType,
    Bbox,
    Caption,
    Cuboid3d,
    DepthAnnotation,
    Ellipse,
    Label,
    LabelCategories,
    Mask,
    Points,
    Polygon,
    PolyLine,
    SuperResolutionAnnotation,
)
from datumaro.components.dataset_base import DatasetItem, IDataset
from datumaro.components.media import Image

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure
    from matplotlib.text import Text

CAPTION_BBOX_PAD = 0.2
DEFAULT_COLOR_CYCLES: List[str] = [
    "#1f77b4",
    "#ff7f0e",
    "#2ca02c",
    "#d62728",
    "#9467bd",
    "#8c564b",
    "#e377c2",
    "#7f7f7f",
    "#bcbd22",
    "#17becf",
]


def _infer_grid_size(length: int, grid_size: Tuple[Optional[int], Optional[int]]):
    nrows, ncols = grid_size

    if nrows is None and ncols is None:
        nrows = ncols = int(math.sqrt(length))

        while nrows * ncols < length:
            nrows += 1
    elif nrows is None and ncols > 0:
        nrows = int(length / ncols)

        while nrows * ncols < length:
            nrows += 1
    elif nrows > 0 and ncols is None:
        ncols = int(length / nrows)

        while nrows * ncols < length:
            ncols += 1

    assert nrows > 0, "nrows should be a positive integer."
    assert ncols > 0, "ncols should be a positive integer."
    assert length <= nrows * ncols, "The number of ids should less then or equal to nrows * ncols."

    return nrows, ncols


class Visualizer:
    def __init__(
        self,
        dataset: IDataset,
        ignored_types: Optional[Iterable[AnnotationType]] = None,
        figsize: Tuple[float, float] = (8, 6),
        color_cycles: Optional[List[str]] = None,
        bbox_linewidth: float = 1.0,
        text_y_offset: float = 1.5,
        alpha: float = 1.0,
        show_plot_title: bool = True,
    ) -> None:
        """
        Visualizer for Datumaro annotations

        Parameters
        ----------
        dataset:
            Datumaro dataset to visualize its items.
        ignored_types:
            Categories of labels. It is used to extract label name by label id.
        figsize:
            Pyplot Figure instance used to draw annotation.
        color_cycles:
            Color cycle corresponding to each label ID.
            If the length of the color cycle is less than the label ID,
            then the label ID exceeding the color cycle length is assgined by the following rule.
            color = color_cycles[label_id % len(color_cycles)]
        bbox_linewidth:
            Line width for Bbox, Polygon and PolyLine annotation
        text_y_offset:
            Offset of y axis for texts.
            The higher value puts the text in the upper place of the annotation.
        alpha:
            Transparency value when drawing annotations. It should be in [0, 1].
            If alpha=0, we do not draw any annotations.
        show_plot_title:
            If True, show the plot title formatted as "ID: {item_id}, Subset: {subset}".
            Otherwise, hide the plot title.
        """
        self.dataset = dataset
        self.figsize = figsize
        self.ignored_types = set(ignored_types) if ignored_types is not None else set()
        self.color_cycles = color_cycles if color_cycles is not None else DEFAULT_COLOR_CYCLES
        self.bbox_linewidth = bbox_linewidth
        self.text_y_offset = text_y_offset

        assert 0.0 <= alpha <= 1.0, "alpha should be in [0, 1]."
        self.alpha = alpha
        self.show_plot_title = show_plot_title

        self._items = [item for item in self.dataset]

    @property
    def draw_only_image(self):
        """
        If self.alpha = 0, we do not overdraw any annotation over the image.
        """
        return self.alpha == 0.0

    def _draw(
        self,
        ann: Annotation,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        """
        Draw annotation according to it's annotation type.

        Parameters
        ----------
        ann:
            Annotation entity to draw.
        label_categories:
            Categories of labels. It is used to extract label name by label id.
        fig:
            Pyplot Figure instance used to draw annotation.
        ax:
            Pyplot Axes instance used to draw annotation.
        context:
            It includes previously drawing history for each annotation type.
            Currently, it is necessary to avoid drawing again over an already drawn place.
            For example, multi label dataset has multiple Labels in a DatasetItem.
            If we don't keep context, it has to draw Label at the same place again and again.
        """
        if ann.type == AnnotationType.label:
            return self._draw_label(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.mask:
            return self._draw_mask(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.points:
            return self._draw_points(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.polygon:
            return self._draw_polygon(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.polyline:
            return self._draw_polygon(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.bbox:
            return self._draw_bbox(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.caption:
            return self._draw_caption(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.cuboid_3d:
            return self._draw_cuboid_3d(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.super_resolution_annotation:
            return self._draw_super_resolution_annotation(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.depth_annotation:
            return self._draw_depth_annotation(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.ellipse:
            return self._draw_ellipse(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.hash_key:
            return None

        # warning instead of raising an error for unsupported annotation types.
        log.warning(f"Ignore unknown ann.type={ann.type}")

    def _get_color(self, ann: Annotation) -> str:
        color = self.color_cycles[ann.label % len(self.color_cycles)]
        return color

    def _sort_by_z_order(self, annotations: List[Annotation]) -> List[Annotation]:
        def _sort_key(ann: Annotation):
            z_order = getattr(ann, "z_order", -1)
            return z_order

        return sorted(annotations, key=_sort_key)

    def get_random_items(self, n_samples: int) -> List[DatasetItem]:
        """Get random samples from the dataset"""
        if n_samples >= len(self.dataset):
            raise ValueError(
                f"n_samples={n_samples} should be less than the dataset size ({len(self.dataset)})."
            )

        # Disable B311: random - used for general random sampling not for security/crypto
        return random.choices(self._items, k=n_samples)  # nosec B311

    @overload
    def vis_gallery(
        self,
        ids: List[str],
        ann_ids: List[int],
        subsets: Union[str, List[str]],
        *,
        grid_size: Tuple[Optional[int], Optional[int]] = (None, None),
    ):
        """Visualize several :class:`DatasetItem` as a gallery.

        If `ann_ids` is given, only draw one annotation matching `ann_id` per item.
        For example, if `ids = ["item_0", "item_1"]` and `ann_ids = [2, 3]`, The item with
        `item.id` = "item_0" will only draw the annotation with `ann.id` = 2 and will not draw the others.

        Parameters
        ----------
        ids
            A list of :class:`DatasetItem`'s ID to visualize
        ann_ids
            A list of :class:`Annotation`'s ID to visualize
        subsets
            A list of :class:`DatasetItem`'s subset name to visualize.
            If a string is given, it is automatically expanded into
            a list up to the length of `ids`.
        grid_size
            Grid size of the gallery plot. If `None`, we automatically infer its size.

        Return
        ------
            :class:`Figure` include visualization plots.
        """
        ...

    @overload
    def vis_gallery(
        self,
        ids: List[str],
        subsets: Union[str, List[str]],
        *,
        grid_size: Tuple[Optional[int], Optional[int]] = (None, None),
    ):
        """Visualize several :class:`DatasetItem` as a gallery

        Parameters
        ----------
        ids
            A list of :class:`DatasetItem`'s ID to visualize
        subsets
            A list of :class:`DatasetItem`'s subset name to visualize.
            If a string is given, it is automatically expanded into
            a list up to the length of `ids`.
        grid_size
            Grid size of the gallery plot. If `None`, we automatically infer its size.

        Return
        ------
            :class:`Figure` include visualization plots.
        """
        ...

    @overload
    def vis_gallery(
        self,
        items: List[DatasetItem],
        *,
        grid_size: Tuple[Optional[int], Optional[int]] = (None, None),
    ):
        """Visualize several :class:`DatasetItem` as a gallery

        Parameters
        ----------
        items
            A list of :class:`DatasetItem` to visualize

        Return
        ------
            :class:`Figure` include visualization plots.
        """
        ...

    def vis_gallery(
        self,
        *inputs,
        grid_size: Tuple[Optional[int], Optional[int]] = (None, None),
    ) -> Figure:
        """Visualize several :class:`DatasetItem` as a gallery"""
        import matplotlib.pyplot as plt

        if len(inputs) == 1:
            (items,) = inputs
            ids = [item.id for item in items]
            subsets = [item.subset for item in items]
            ann_ids = [None for _ in items]
        elif len(inputs) == 2:
            ids, subsets = inputs
            items = None
            ann_ids = [None for _ in ids]
        elif len(inputs) == 3:
            ids, ann_ids, subsets = inputs

        if isinstance(subsets, str):
            subsets = [subsets] * len(ids)  # expand it to have len(ids)

        assert (
            len(ids) == len(ann_ids) == len(subsets)
        ), "ids, ann_ids, subset should have the same length"

        nrows, ncols = _infer_grid_size(len(ids), grid_size)
        fig, axs = plt.subplots(nrows, ncols, figsize=self.figsize)

        assert len(ids) == len(
            subsets
        ), "If subset is a list, it should have the same length as ids."

        for item_id, subset, ann_id, ax in zip(ids, subsets, ann_ids, axs.flatten()):
            self.vis_one_sample(item_id, subset, ann_id=ann_id, ax=ax)

        return fig

    @overload
    def vis_one_sample(
        self,
        item_id: str,
        subset: str,
        *,
        ann_id: Optional[int] = None,
        ax: Optional[Axes] = None,
    ) -> Figure:
        """Visualize one :class:`DatasetItem`

        Parameters
        ----------
        item_id
            ID of :class:`DatasetItem` to visualize
        subset
            Subset name of :class:`DatasetItem` to visualize
        ann_id
            If not `None`, only draw an annotation which has `id=ann_id`.
        ax
            If not `None`, draw on `ax` instead of creating a new one

        Return
        ------
            :class:`Figure` include visualization plot of the :class:`DatasetItem`.
        """
        ...

    @overload
    def vis_one_sample(
        self,
        item: DatasetItem,
        *,
        ann_id: Optional[int] = None,
        ax: Optional[Axes] = None,
    ) -> Figure:
        """Visualize one :class:`DatasetItem`

        Parameters
        ----------
        item
            :class:`DatasetItem` to visualize
        ann_id
            If not `None`, only draw an annotation which has `id=ann_id`.
        ax
            If not `None`, draw on `ax` instead of creating a new one

        Return
        ------
            :class:`Figure` include visualization plot of the :class:`DatasetItem`.
        """
        ...

    def vis_one_sample(
        self,
        *inputs,
        ann_id: Optional[int] = None,
        ax: Optional[Axes] = None,
    ) -> Figure:
        """Visualize one dataset item"""
        import matplotlib.pyplot as plt

        if len(inputs) == 1:
            item_id, subset = None, None
            (item,) = inputs

        elif len(inputs) == 2:
            item_id, subset = inputs
            item = None

        if ax is None:
            fig = plt.figure(figsize=self.figsize)
            ax = plt.gca()
        else:
            fig = ax.get_figure()
            plt.sca(ax)

        def _parse_inputs(
            item_id: Optional[str], subset: Optional[str], item: Optional[DatasetItem]
        ) -> Tuple[str, str]:
            if item_id is not None and subset is not None:
                return item_id, subset
            elif item is not None:
                item_id = item.id
                subset = item.subset
                return item_id, subset
            raise ValueError(
                f"item_id={item_id}, subset={subset}, and item={item} is an invalid input."
            )

        item_id, subset = _parse_inputs(item_id, subset, item)
        item: DatasetItem = self.dataset.get(item_id, subset)

        assert item is not None, f"Cannot find id={item_id}, subset={subset}"
        assert (
            item is not Image
        ), f"Media type should be Image, Current media type={type(item.media)}"

        img = item.media.data.astype(np.uint8)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        ax.imshow(img)

        if self.show_plot_title:
            self._plot_title(ax, item_id, subset)

        ax.set_axis_off()

        if self.draw_only_image:
            return fig

        annotations = self._sort_by_z_order(item.annotations)
        categories = self.dataset.categories()
        label_categories = (
            self.dataset.categories()[AnnotationType.label]
            if AnnotationType.label in categories
            else None
        )

        context = defaultdict(list)
        for ann in annotations:
            if ann.type in self.ignored_types:
                ignore_type = AnnotationType(ann.type).name
                msg = f"{ignore_type} in self.ignored_types. Skip it."
                warnings.warn(msg)
                continue

            if ann_id is None or ann_id == ann.id:
                self._draw(ann, label_categories, fig, ax, context[ann.type])

        return fig

    def _plot_title(self, ax: Axes, item_id: str, subset: str) -> None:
        width = ax.transAxes.transform_point((1, 0))[0] - ax.transAxes.transform_point((0, 0))[0]
        text = ax.set_title(f"ID: {item_id}, Subset: {subset}", loc="center", wrap=True)
        text.__get_wrapped_text = text._get_wrapped_text

        def _get_wrapped_text():
            wrapped_text = text.__get_wrapped_text()
            text._text = wrapped_text
            return wrapped_text

        text._get_wrapped_text = _get_wrapped_text
        text._get_wrap_line_width = lambda: width

    def _draw_label(
        self,
        ann: Label,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        label_text = label_categories[ann.label].name if label_categories is not None else ann.label
        color = self._get_color(ann)

        if len(context) == 0:
            x, y = 0.01, 0.99
        else:
            # Draw below the previously drawn label.
            text: Text = context[-1]
            # We can know the position of text bbox only if drawing it actually.
            # https://stackoverflow.com/a/41271773/16880031
            fig.canvas.draw()
            bbox = text.get_window_extent()
            bbox = ax.transAxes.inverted().transform_bbox(bbox)
            x, y = 0.01, bbox.y0

        text = ax.text(x, y, label_text, ha="left", va="top", color=color, transform=ax.transAxes)
        context.append(text)

    def _draw_mask(
        self,
        ann: Mask,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        h, w = ann.image.shape
        source = ann.image
        if source.dtype != bool:
            warnings.warn(
                f"Mask should has dtype == bool, but its dtype == {source.dtype}. "
                "Try to change it to bool dtype."
            )
            source = source.astype(bool)

        mask_map = np.zeros((h, w, 4), dtype=np.uint8)
        color = self._get_color(ann)
        rgba_color = (*ImageColor.getcolor(color, "RGB"), 0.0)
        mask_map[source] = rgba_color
        mask_map[source, 3] = int(255 * self.alpha)

        ax.imshow(mask_map)

    def _draw_points(
        self,
        ann: Points,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        label_text = label_categories[ann.label].name if label_categories is not None else ann.label
        color = self._get_color(ann)
        points = np.array(ann.points)
        n_points = len(points) // 2
        points = points.reshape(n_points, 2)
        visible = [viz == Points.Visibility.visible for viz in ann.visibility]
        points = points[visible]

        ax.scatter(points[:, 0], points[:, 1], color=color)

        if len(points) > 0:
            x, y, _, _ = ann.get_bbox()
            ax.text(x, y - self.text_y_offset, label_text, color=color)

    def _draw_polygon(
        self,
        ann: Union[Polygon, PolyLine],
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        import matplotlib.patches as patches

        label_text = label_categories[ann.label].name if label_categories is not None else ann.label
        color = self._get_color(ann)
        points = np.array(ann.points)
        n_points = len(points) // 2
        points = points.reshape(n_points, 2)

        polyline = patches.Polygon(
            points,
            fill=False,
            linewidth=self.bbox_linewidth,
            edgecolor=color,
        )
        ax.add_patch(polyline)

        if isinstance(ann, Polygon):
            polygon = patches.Polygon(
                points,
                fill=True,
                facecolor=color if isinstance(ann, Polygon) else "none",
                alpha=self.alpha,
            )
            ax.add_patch(polygon)

        x, y, _, _ = ann.get_bbox()
        ax.text(x, y - self.text_y_offset, label_text, color=color)

    def _draw_bbox(
        self,
        ann: Bbox,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        import matplotlib.patches as patches

        label_text = label_categories[ann.label].name if label_categories is not None else ann.label
        color = self._get_color(ann)
        rect = patches.Rectangle(
            (ann.x, ann.y),
            ann.w,
            ann.h,
            linewidth=self.bbox_linewidth,
            edgecolor=color,
            facecolor="none",
        )
        ax.text(ann.x, ann.y - self.text_y_offset, label_text, color=color)
        ax.add_patch(rect)

    def _draw_caption(
        self,
        ann: Caption,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        if len(context) == 0:
            x, y = 0.5, 0.01
        else:
            # Draw on the top of the previously drawn caption.
            text: Text = context[-1]
            bbox = text.get_bbox_patch()
            # We can know the position of text bbox only if drawing it actually.
            # https://stackoverflow.com/a/41271773/16880031
            fig.canvas.draw()
            drawed_bbox = bbox.get_bbox()
            x, y = bbox.get_transform().transform(
                [0, (1.0 + 2 * CAPTION_BBOX_PAD) * drawed_bbox.height]
            )
            x, y = ax.transAxes.inverted().transform([x, y])
            x, y = 0.5, y

        width = ax.transAxes.transform_point((1, 0))[0] - ax.transAxes.transform_point((0, 0))[0]

        text = ax.text(
            x,
            y,
            ann.caption,
            ha="center",
            va="bottom",
            wrap=True,
            transform=ax.transAxes,
            bbox={"facecolor": "white", "alpha": self.alpha},
        )
        text.get_bbox_patch().set_boxstyle(f"Round,pad={CAPTION_BBOX_PAD}")
        text._get_wrap_line_width = lambda: width
        context.append(text)

    def _draw_cuboid_3d(
        self,
        ann: Cuboid3d,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        raise NotImplementedError(f"{ann.type} is not implemented yet.")

    def _draw_super_resolution_annotation(
        self,
        ann: SuperResolutionAnnotation,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        assert (
            len(context) == 0
        ), "It cannot visualize more than one SuperResolutionAnnotation per item."

        warnings.warn(
            "SuperResolutionAnnotation overdraws the high-resolution image over the original image. "
            "If you want to see the original image, set alpha=0."
        )

        hi_res_img = ann.image.data
        im = ax.imshow(hi_res_img, alpha=self.alpha)
        context.append(im)

    def _draw_depth_annotation(
        self,
        ann: DepthAnnotation,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        from mpl_toolkits.axes_grid1 import make_axes_locatable

        assert len(context) == 0, "It cannot visualize more than one DepthAnnotation per item."

        depth = ann.image.data

        im = ax.imshow(depth, alpha=self.alpha)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)

        fig.colorbar(im, cax)
        context.append(im)

    def _draw_ellipse(
        self,
        ann: Ellipse,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        import matplotlib.patches as patches

        label_text = label_categories[ann.label].name if label_categories is not None else ann.label
        color = self._get_color(ann)
        ellipse = patches.Ellipse(
            xy=(ann.c_x, ann.c_y),
            width=ann.w,
            height=ann.h,
            linewidth=self.bbox_linewidth,
            edgecolor=color,
            facecolor="none",
        )
        ax.text(ann.x1, ann.y1 - self.text_y_offset, label_text, color=color)
        ax.add_patch(ellipse)
